"""
Mask R-CNN
Base Configurations class.

Copyright (c) 2017 Matterport, Inc.
Licensed under the MIT License (see LICENSE for details)
Written by Waleed Abdulla
"""

import os
import logging
import json
import datetime
import numpy as np
import skimage.io
#from imgaug import augmenters as iaa
import instantdl.segmentation.visualizeRCNN as visualize
import matplotlib
import logging
# Agg backend runs without a display
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import instantdl.segmentation.utilsRCNN as utils
from instantdl.data_generator.data_generator import import_image

# Base Configuration Class
# Don't use this class directly. Instead, sub-class it and override
# the configurations you need to change.

class Config(object):
    """Base configuration class. For custom configurations, create a
    sub-class that inherits from this one and override properties
    that need to be changed.
    """
    # Name the configurations. For example, 'COCO', 'Experiment 3', ...etc.
    # Useful if your code needs to do things differently depending on which
    # experiment is running.
    NAME = "Images"  # Override in sub-classes

    # NUMBER OF GPUs to use. When using only a CPU, this needs to be set to 1.
    GPU_COUNT = 1

    # Number of images to train with on each GPU. A 12GB GPU can typically
    # handle 2 images of 1024x1024px.
    # Adjust based on your GPU memory and image sizes. Use the highest
    # number that your GPU can handle for best performance.
    IMAGES_PER_GPU = 1

    # Number of training steps per epoch
    # This doesn't need to match the size of the training set. Tensorboard
    # updates are saved at the end of each epoch, so setting this to a
    # smaller number means getting more frequent TensorBoard updates.
    # Validation stats are also calculated at each epoch end and they
    # might take a while, so don't set this too small to avoid spending
    # a lot of time on validation stats.
    STEPS_PER_EPOCH = 100

    # Number of validation steps to run at the end of every training epoch.
    # A bigger number improves accuracy of validation stats, but slows
    # down the training.
    VALIDATION_STEPS = 10

    # Backbone network architecture
    # Supported values are: resnet50, resnet101.
    # You can also provide a callable that should have the signature
    # of model.resnet_graph. If you do so, you need to supply a callable
    # to COMPUTE_BACKBONE_SHAPE as well
    BACKBONE = "resnet50"

    # Only useful if you supply a callable to BACKBONE. Should compute
    # the shape of each layer of the FPN Pyramid.
    # See model.compute_backbone_shapes
    COMPUTE_BACKBONE_SHAPE = None

    # The strides of each layer of the FPN Pyramid. These values
    # are based on a Resnet101 backbone.
    BACKBONE_STRIDES = [4, 8, 16, 32, 64]

    # Size of the fully-connected layers in the classification graph
    FPN_CLASSIF_FC_LAYERS_SIZE = 1024

    # Size of the top-down layers used to build the feature pyramid
    TOP_DOWN_PYRAMID_SIZE = 256

    # Number of classification classes (including background)
    NUM_CLASSES = 1 + 1 # Override in sub-classes

    # Length of square anchor side in pixels
    RPN_ANCHOR_SCALES = (8, 16, 32, 64, 256) #until 256

    # Ratios of anchors at each cell (width/height)
    # A value of 1 represents a square anchor, and 0.5 is a wide anchor
    RPN_ANCHOR_RATIOS = [0.5, 1, 2]

    # Anchor stride
    # If 1 then anchors are created for each cell in the backbone feature map.
    # If 2, then anchors are created for every other cell, and so on.
    RPN_ANCHOR_STRIDE = 1

    # Non-max suppression threshold to filter RPN proposals.
    # You can increase this during training to generate more propsals.
    RPN_NMS_THRESHOLD = 0.8

    # How many anchors per image to use for RPN training
    RPN_TRAIN_ANCHORS_PER_IMAGE = 128

    # ROIs kept after tf.nn.top_k and before non-maximum suppression
    PRE_NMS_LIMIT = 6000

    # ROIs kept after non-maximum suppression (training and inference)
    POST_NMS_ROIS_TRAINING = 1000
    POST_NMS_ROIS_INFERENCE = 2000

    # If enabled, resizes instance masks to a smaller size to reduce
    # memory load. Recommended when using high-resolution images.
    USE_MINI_MASK = True
    MINI_MASK_SHAPE = (56, 56) # (height, width) of the mini-mask

    # Input image resizing
    # Generally, use the "square" resizing mode for training and predicting
    # and it should work well in most cases. In this mode, images are scaled
    # up such that the small side is = IMAGE_MIN_DIM, but ensuring that the
    # scaling doesn't make the long side > IMAGE_MAX_DIM. Then the image is
    # padded with zeros to make it a square so multiple images can be put
    # in one batch.
    # Available resizing modes:
    # none:   No resizing or padding. Return the image unchanged.
    # square: Resize and pad with zeros to get a square image
    #         of size [max_dim, max_dim].
    # pad64:  Pads width and height with zeros to make them multiples of 64.
    #         If IMAGE_MIN_DIM or IMAGE_MIN_SCALE are not None, then it scales
    #         up before padding. IMAGE_MAX_DIM is ignored in this mode.
    #         The multiple of 64 is needed to ensure smooth scaling of feature
    #         maps up and down the 6 levels of the FPN pyramid (2**6=64).
    # crop:   Picks random crops from the image. First, scales the image based
    #         on IMAGE_MIN_DIM and IMAGE_MIN_SCALE, then picks a random crop of
    #         size IMAGE_MIN_DIM x IMAGE_MIN_DIM. Can be used in training only.
    #         IMAGE_MAX_DIM is not used in this mode.
    IMAGE_RESIZE_MODE = "crop"
    IMAGE_MIN_DIM = 256
    IMAGE_MAX_DIM = 512

    # Minimum scaling ratio. Checked after MIN_IMAGE_DIM and can force further
    # up scaling. For example, if set to 2 then images are scaled up to double
    # the width and height, or more, even if MIN_IMAGE_DIM doesn't require it.
    # However, in 'square' mode, it can be overruled by IMAGE_MAX_DIM.
    IMAGE_MIN_SCALE = 0
    # Number of color channels per image. RGB = 3, grayscale = 1, RGB-D = 4
    # Changing this requires other changes in the code. See the WIKI for more
    # details: https://github.com/matterport/Mask_RCNN/wiki
    IMAGE_CHANNEL_COUNT = 3

    # Image mean (RGB)
    MEAN_PIXEL = np.array([43.53, 39.56, 48.22])

    # Number of ROIs per image to feed to classifier/mask heads
    # The Mask RCNN paper uses 512 but often the RPN doesn't generate
    # enough positive proposals to fill this and keep a positive:negative
    # ratio of 1:3. You can increase the number of proposals by adjusting
    # the RPN NMS threshold.
    TRAIN_ROIS_PER_IMAGE = 200

    # Percent of positive ROIs used to train classifier/mask heads
    ROI_POSITIVE_RATIO = 0.33

    # Pooled ROIs
    POOL_SIZE = 7
    MASK_POOL_SIZE = 14

    # Shape of output mask
    # To change this you also need to change the neural network mask branch
    MASK_SHAPE = [28, 28]

    # Maximum number of ground truth instances to use in one image
    MAX_GT_INSTANCES = 200

    # Bounding box refinement standard deviation for RPN and final detections.
    RPN_BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])
    BBOX_STD_DEV = np.array([0.1, 0.1, 0.2, 0.2])

    # Max number of final detections
    DETECTION_MAX_INSTANCES = 400

    # Minimum probability value to accept a detected instance
    # ROIs below this threshold are skipped
    DETECTION_MIN_CONFIDENCE = 0.85

    # Non-maximum suppression threshold for detection
    #Overlap of detections: close to 1 means a lot of overlap, close to 0, no overlap
    DETECTION_NMS_THRESHOLD = 0.3

    # Learning rate and momentum
    # The Mask RCNN paper uses lr=0.02, but on TensorFlow it causes
    # weights to explode. Likely due to differences in optimizer
    # implementation.
    LEARNING_RATE =  0.0002 #0.0001
    LEARNING_MOMENTUM = 0.9

    # Weight decay regularization
    WEIGHT_DECAY = 0.0001

    # Loss weights for more precise optimization.
    # Can be used for R-CNN training setup.
    LOSS_WEIGHTS = {
        "rpn_class_loss": 1.,
        "rpn_bbox_loss": 1.,
        "mrcnn_class_loss": 1.,
        "mrcnn_bbox_loss": 1.,
        "mrcnn_mask_loss": 1.
    }

    # Use RPN ROIs or externally generated ROIs for training
    # Keep this True for most situations. Set to False if you want to train
    # the head branches on ROI generated by code rather than the ROIs from
    # the RPN. For example, to debug the classifier head without having to
    # train the RPN.
    USE_RPN_ROIS = True

    # Train or freeze batch normalization layers
    #     None: Train BN layers. This is the normal mode
    #     False: Freeze BN layers. Good when using a small batch size
    #     True: (don't use). Set layer in training mode even when predicting
    TRAIN_BN = False  # Defaulting to False since batch size is often small

    # Gradient norm clipping
    GRADIENT_CLIP_NORM = 5.0

    def __init__(self):
        """Set values of computed attributes."""
        # Effective batch size
        self.BATCH_SIZE = self.IMAGES_PER_GPU * self.GPU_COUNT

        # Input image size
        if self.IMAGE_RESIZE_MODE == "crop":
            self.IMAGE_SHAPE = np.array([self.IMAGE_MIN_DIM, self.IMAGE_MIN_DIM,
                                         self.IMAGE_CHANNEL_COUNT])
        else:
            self.IMAGE_SHAPE = np.array([self.IMAGE_MAX_DIM, self.IMAGE_MAX_DIM,
                                         self.IMAGE_CHANNEL_COUNT])

        # Image meta data length
        # See compose_image_meta() for details
        self.IMAGE_META_SIZE = 1 + 3 + 3 + 4 + 1 + self.NUM_CLASSES

    def display(self):
        """Display Configuration values."""
        logging.info("\nConfigurations:")
        for a in dir(self):
            if not a.startswith("__") and not callable(getattr(self, a)):
                logging.info("{:30} {}".format(a, getattr(self, a)))
        logging.info("\n")


class RCNNConfig(Config):

    """Configuration for training on the nucleus segmentation dataset."""
    # Give the configuration a recognizable name
    NAME = "Image"

    # Number of classes (including background)
    NUM_CLASSES = 1 + 1  # Background + nucleus

    # Number of training and validation steps per epoch
    STEPS_PER_EPOCH = 100 # (657 - len(VAL_IMAGE_IDS)) // IMAGES_PER_GPU
    VALIDATION_STEPS = 10 # max(1, len(VAL_IMAGE_IDS) // IMAGES_PER_GPU)

    # Backbone network architecture
    # Supported values are: resnet50, resnet101
    BACKBONE = "resnet50"

    classdefs = ("Image", 1, "Detection")

class RCNNInferenceConfig(RCNNConfig):
    # Set batch size to 1 to run one image at a time
    GPU_COUNT = 1
    IMAGES_PER_GPU = 1
    # Don't resize images for inferencing
    IMAGE_RESIZE_MODE = "pad64"
    # Non-max suppression threshold to filter RPN proposals.
    # You can increase this during training to generate more propsals.
    RPN_NMS_THRESHOLD = 0.7 #0.7


############################################################
#  Dataset
############################################################

class NucleusDataset(utils.Dataset):

    def load_dataset(self, dataset_dir, subset, VAL_IMAGE_IDS):
        """Load a subset of the nuclei dataset.

        dataset_dir: Root directory of the dataset
        subset: Subset to load. Either the name of the sub-directory,
                such as stage1_train, stage1_test, ...etc. or, one of:
                * train: stage1_train excluding validation images
                * val: validation images from VAL_IMAGE_IDS
        """
        # Add classes. We have one class.
        # Naming the dataset nucleus, and the class nucleus
        self.add_class("image", 1, "Inside Cell")

        subset_dir = "train" if subset in ["train", "val"] else subset
        dataset_dir = os.path.join(dataset_dir, subset_dir)
        logging.info("Datasetdir", dataset_dir)
        if subset == "val":
            image_ids = VAL_IMAGE_IDS
        else:
            # Get image ids from directory names
            image_ids = next(os.walk(dataset_dir))[1]
            if subset == "train":
                image_ids = list(set(image_ids) - set(VAL_IMAGE_IDS))

        # Add images
        for image_id in image_ids:
            self.add_image(
                "image",
                image_id=image_id,
                path=os.path.join(dataset_dir, image_id, "image/{}.png".format(image_id)))

    def load_mask(self, image_id):
        """Generate instance masks for an image.
       Returns:
        masks: A bool array of shape [height, width, instance count] with
            one mask per instance.
        class_ids: a 1D array of class IDs of the instance masks.
        """
        info = self.image_info[image_id]
        # Get mask directory from image path
        mask_dir = os.path.join(os.path.dirname(os.path.dirname(info['path'])), "mask")

        # Read mask files from .png image
        mask = []
        for f in next(os.walk(mask_dir))[2]:
            #if f.endswith(".png"):
            m = import_image(os.path.join(mask_dir, f)).astype(np.bool)
            #m = skimage.io.imread(os.path.join(mask_dir, f)).astype(np.bool)
            if np.shape(m)[-1] > 1:
                from skimage.color import rgb2gray
                m = rgb2gray(m)
            mask.append(m)

        mask = np.stack(mask, axis=-1)
        #logging.info("loading masks", np.shape(mask))
        # Return mask, and array of class IDs of each instance. Since we have
        # one class ID, we return an array of ones
        return mask, np.ones([mask.shape[-1]], dtype=np.int32)

    def image_reference(self, image_id):
        """Return the path of the image."""
        info = self.image_info[image_id]
        if info["source"] == "Inner Cell":
            return info["id"]
        else:
            super(self.__class__, self).image_reference(image_id)


############################################################
#  Training
############################################################

def train(model, dataset_dir, subset, VAL_IMAGE_IDS, epochs, custom_callbacks):
    """Train the model."""
    # Training dataset.
    dataset_train = NucleusDataset()

    dataset_train.load_dataset(dataset_dir, subset, VAL_IMAGE_IDS)
    dataset_train.prepare()

    # Validation dataset
    dataset_val = NucleusDataset()
    dataset_val.load_dataset(dataset_dir, "val", VAL_IMAGE_IDS)
    dataset_val.prepare()

    logging.info("Train all layers")
    model.train(dataset_train, dataset_val,
                learning_rate=RCNNConfig.LEARNING_RATE,
                epochs=epochs,
                custom_callbacks = custom_callbacks,
                augmentation=None,
                layers='all')


############################################################
#  RLE Encoding
############################################################

def rle_encode(mask):
    """Encodes a mask in Run Length Encoding (RLE).
    Returns a string of space-separated values.
    """
    assert mask.ndim == 2, "Mask must be of shape [Height, Width]"
    # Flatten it column wise
    m = mask.T.flatten()
    # Compute gradient. Equals 1 or -1 at transition points
    g = np.diff(np.concatenate([[0], m, [0]]), n=1)
    # 1-based indicies of transition points (where gradient != 0)
    rle = np.where(g != 0)[0].reshape([-1, 2]) + 1
    # Convert second index in each pair to lenth
    rle[:, 1] = rle[:, 1] - rle[:, 0]
    return " ".join(map(str, rle.flatten()))


def rle_decode(rle, shape):
    """Decodes an RLE encoded list of space separated
    numbers and returns a binary mask."""
    rle = list(map(int, rle.split()))
    rle = np.array(rle, dtype=np.int32).reshape([-1, 2])
    rle[:, 1] += rle[:, 0]
    rle -= 1
    mask = np.zeros([shape[0] * shape[1]], np.bool)
    for s, e in rle:
        assert 0 <= s < mask.shape[0]
        assert 1 <= e <= mask.shape[0], "shape: {}  s {}  e {}".format(shape, s, e)
        mask[s:e] = 1
    # Reshape and transpose
    mask = mask.reshape([shape[1], shape[0]]).T
    return mask


def mask_to_rle(image_id, mask, scores):
    "Encodes instance masks to submission format."
    assert mask.ndim == 3, "Mask must be [H, W, count]"
    # If mask is empty, return line with image ID only
    if mask.shape[-1] == 0:
        return "{},".format(image_id)
    # Remove mask overlaps
    # Multiply each instance mask by its score order
    # then take the maximum across the last dimension
    order = np.argsort(scores)[::-1] + 1  # 1-based descending
    mask = np.max(mask * np.reshape(order, [1, 1, -1]), -1)
    # Loop over instance masks
    lines = []
    for o in order:
        m = np.where(mask == o, 1, 0)
        # Skip if empty
        if m.sum() == 0.0:
            continue
        rle = rle_encode(m)
        lines.append("{}, {}".format(image_id, rle))
    return "\n".join(lines)


############################################################
#  Detection
############################################################

def detect(model, dataset_dir, subset, RESULTS_DIR, VAL_IMAGE_IDS):
    """Run detection on images in the given directory."""
    logging.info("Running on {}".format(dataset_dir))

    # Create directory
    if not os.path.exists(RESULTS_DIR):
        os.makedirs(RESULTS_DIR)
    #submit_dir = "submit_{:%Y%m%dT%H%M%S}".format(datetime.datetime.now()) + "/"
    #submit_dir = os.path.join(RESULTS_DIR)
    submit_dir = RESULTS_DIR
    os.makedirs("./" + (submit_dir + "/logs"), exist_ok=True)

    # Read dataset
    dataset = NucleusDataset()
    VAL_IMAGE_IDS = []
    dataset.load_dataset(dataset_dir, subset, VAL_IMAGE_IDS)
    dataset.prepare()
    # Load over images
    submission = []
    for image_id in dataset.image_ids:
        logging.info("Test image id", image_id)
        # Load image and run detection
        image = dataset.load_image(image_id)
        # Detect objects
        r = model.detect([image], verbose=0)[0]
        # Encode image to RLE. Returns a string of multiple lines
        source_id = dataset.image_info[image_id]["id"]
        #rle = mask_to_rle(source_id, r["masks"], r["scores"])
        #submission.append(rle)

        logging.info("detection masks and scores", np.shape(r["masks"]), np.shape(r["scores"]))
        np.save(submit_dir + dataset.image_info[image_id]["id"], r["masks"])

        #TODO: Uncertanty Prediction with MC Dropout:



        # Save image with masks
        visualize.display_instances(
            image, r['rois'], r['masks'], r['class_ids'],
            dataset.class_names, r['scores'],
            show_bbox=False, show_mask=False,
            title="Predictions")
        plt.savefig("{}/{}.png".format(submit_dir, dataset.image_info[image_id]["id"]))


    # Save to csv file
    #submission = "ImageId,EncodedPixels\n" + "\n".join(submission)
    #file_path = os.path.join(submit_dir, "submit.csv")
    #with open(file_path, "w") as f:
    #    f.write(submission)
    #logging.info("Saved to ", submit_dir)




'''
InstantDL
Written by Dominik Waibel and Ali Boushehri

In this file the functions are started to train and test the networks
'''

from instantdl.utils import *
from instantdl.segmentation.UNet_models import UNetBuilder

class Regression(object):
    def __init__(   self,
                    use_algorithm,
                    path,
                    pretrained_weights = None,
                    batchsize = 2,
                    iterations_over_dataset = 100,
                    data_gen_args = None,
                    loss_function = "mse",
                    num_classes = 1,
                    image_size = None,
                    seeds=False,
                    calculate_uncertainty = False,
                    evaluation = True):

        self.use_algorithm = "Regression"
        self.path = path
        self.pretrained_weights = pretrained_weights
        self.batchsize = batchsize
        self.iterations_over_dataset = iterations_over_dataset
        self.loss_function = loss_function
        self.num_classes = num_classes
        self.image_size = image_size
        self.calculate_uncertainty = calculate_uncertainty
        
        if data_gen_args is None:
            self.data_gen_args = dict()
        else:
            self.data_gen_args = data_gen_args
        self.evaluation = evaluation
    
    def data_prepration(self): 
        '''
        Get the number of input images and their shape
        If the last image dimension,. which should contain the channel information (1 or 3) is not existing e.g. for 
        (512,512) add a 1 as the channel number.
        '''
        if self.image_size == False or self.image_size == None:
            Training_Input_shape, num_channels, Input_image_shape = get_input_image_sizes(self.path, self.use_algorithm)
        else:
            Training_Input_shape = self.image_size
            num_channels = int(self.image_size[-1])
            data_path = self.path + '/train'
            img_file = os.listdir(data_path + "/image/")[0]
            Input_image_shape = np.array(np.shape(np.array(import_image(data_path + "/image/" + img_file))))

        ''' 
        Check if the 2D or 3D Pipeline is needed
        '''
        if len(Training_Input_shape[:-1]) == 3:
            data_dimensions = 3
        if len(Training_Input_shape[:-1]) == 2:
            data_dimensions = 2

        logging.info("Image dimensions are: %s D" % data_dimensions )

        Folders = ["image", "image1", "image2", "image3", "image4", "image5", "image6", "image7"]
        number_input_images = len([element for element in os.listdir(self.path + "/train/") if element in Folders])
        network_input_size = np.array(Training_Input_shape)
        network_input_size[-1] = int(Training_Input_shape[-1]) * number_input_images
        network_input_size = tuple(network_input_size)
        logging.info("Number of input folders is: %s" % number_input_images)
        logging.info("UNet input shape %s" % (network_input_size,))

        '''
        Import filenames and split them into train and validation set according to 
        the variable -validation_split = 20%
        '''
        data_path = self.path + '/train'
        train_image_files, val_image_files = training_validation_data_split(data_path)

        steps_per_epoch = int(len(train_image_files)/self.batchsize)

        self.epochs = self.iterations_over_dataset
        logging.info("Making: %s steps per Epoch" % steps_per_epoch)
        return [Training_Input_shape, num_channels, network_input_size, Input_image_shape,
                        data_path, train_image_files, val_image_files, steps_per_epoch, data_dimensions,val_image_files]
        #TODO: Returning val_image_files twice



    def data_generator(self, data_path, Training_Input_shape, num_channels, train_image_files, data_dimensions, val_image_files):
        '''
        Prepare data as a Training and Validation set
        Args:
            data_path: Path to folder containing the dataset
            Training_Input_shape: Shape of the input images in the train folder
            num_channels: Number of channels (e.g.: 3 for RGB)
            train_image_files: List of filenames contained in the train set
            val_image_files: List of filenames contained in the validation set
            data_dimensions: Image dimensions
        return:
            Two data generators (train & validation) and the number of channels of the groundtruth (label)
        '''


        img_file_label_name = os.listdir(data_path + "/groundtruth/")[0]
        logging.info("img_file_label_name: %s" % img_file_label_name)
        Training_Input_shape_label = np.shape(np.array(import_image(data_path + "/groundtruth/" + img_file_label_name)))
        num_channels_label = Training_Input_shape_label[-1]
        if all([num_channels_label != 1, num_channels_label != 3]):
            num_channels_label = 1

        if self.use_algorithm == "SemanticSegmentation":
            self.data_gen_args["binarize_mask"] = True

        TrainingDataGenerator = training_data_generator(Training_Input_shape,
                                                            self.batchsize, num_channels,
                                                            num_channels_label,
                                                            train_image_files,
                                                            self.data_gen_args,
                                                            data_dimensions,
                                                            data_path,
                                                            self.use_algorithm)
        ValidationDataGenerator = training_data_generator(Training_Input_shape,
                                                              self.batchsize, num_channels,
                                                              num_channels_label,
                                                              val_image_files,
                                                              self.data_gen_args,
                                                              data_dimensions,
                                                              data_path,
                                                              self.use_algorithm)
        return TrainingDataGenerator, ValidationDataGenerator,num_channels_label
    
    def load_model(self, network_input_size,data_dimensions,num_channels_label ):
        '''
        Build a 2D or 3D U-Net model and initialize it with pretrained or random weights
        Args:
            network_input_size: Dimensions of one input image (e.g. 128,128,3)
            data_dimensions: Dimensions of the data (e.g. 3)
            num_channels_label: Number of channels of the groundtruth (e.g.: 3 for RGB)
        returns:
            A 2D or 3D UNet model
        '''
        if self.pretrained_weights == False:
            self.pretrained_weights = None
        if data_dimensions == 3:
            logging.info("Using 3D UNet")
            model = UNetBuilder.unet3D(self.pretrained_weights, network_input_size[-1], num_channels_label, self.num_classes, self.loss_function, Dropout_On = True)
        else:
            logging.info("Using 2D UNet")
            model = UNetBuilder.unet2D(self.pretrained_weights, network_input_size[-1], num_channels_label, self.num_classes, self.loss_function, Dropout_On = True)

        logging.info(model.summary())
        return model

    def train_model(self, model,TrainingDataGenerator,ValidationDataGenerator , steps_per_epoch, val_image_files ):
        '''
        Set Model callbacks such as: 
        - Early stopping (after the validation loss has not improved for 25 epochs
        - Checkpoints: Save model after each epoch if the validation loss has improved 
        - Tensorboard: Monitor training live with tensorboard. Start tensorboard in terminal with: tensorboard --logdir=/path_to/logs 
        Args:
            model: The initialized U-Net model
            TrainingDataGenerator: The train data generator
            ValidationDataGenerator: The validation data generator
            steps_per_epoch: The number of train steps in one epoch
            val_image_files: List of validation files
        returns:
            The trained model and the checkpoint file path
        '''

        early_stopping = EarlyStopping(monitor='val_loss', patience=5, mode='auto', verbose=0)
        datasetname = self.path.rsplit("/",1)[1]
        checkpoint_filepath = (self.path + "/logs" + "/pretrained_weights" + datasetname + ".hdf5") #.{epoch:02d}.hdf5")
        os.makedirs((self.path + "/logs"), exist_ok=True)
        model_checkpoint = ModelCheckpoint(checkpoint_filepath, monitor=('val_loss'), verbose=1, save_best_only=True)

        tensorboard = TensorBoard(log_dir = self.path + "logs/" + "/" + format(time.time())) #, update_freq='batch')
        logging.info("Tensorboard log is created at: logs/  it can be opend using tensorboard --logdir=logs for a terminal in the Project folder")
        callbacks_list = [model_checkpoint, tensorboard, early_stopping]

        '''
        Train the model given the initialized model and the data from the data generator
        '''
        model.fit_generator(TrainingDataGenerator,
                                steps_per_epoch=steps_per_epoch,
                                validation_data=ValidationDataGenerator,
                                validation_steps=len(val_image_files),
                                max_queue_size=50,
                                epochs=self.epochs,
                                callbacks = callbacks_list,
                                use_multiprocessing=True)
        logging.info('finished Model.fit_generator')
        return model, checkpoint_filepath

    def test_set_evaluation(self, model, Training_Input_shape, num_channels,Input_image_shape):
        '''
        Evalute the model on the testset
        Args:
            model: the trained or initialized model
            Training_Input_shape: The dimensions of the input data
            num_channels: Number of channels (e.g.: 3 for RGB)
            Input_image_shape: The shape of the input images
        returns: the results of the tested images, a list of filenames of the testset, the number of images tested
        '''
        test_image_files = os.listdir(os.path.join(self.path + "/test/image"))
        num_test_img = int(len(os.listdir(self.path + "/test/image")))

        '''
        Initialize the testset generator
        '''
        testGene = testGenerator(Training_Input_shape, self.path, num_channels, test_image_files, self.use_algorithm)
        logging.info('finished testGene')
        results = model.predict_generator(testGene, steps=num_test_img, use_multiprocessing=False, verbose=1)
        #logging.info("results"), np.shape(results))
        logging.info('finished model.predict_generator')


        '''
        Save the models prediction on the testset by printing the predictions 
        as images to the results folder in the project path
        '''
        saveResult(self.path + "/results/", test_image_files, results, Input_image_shape)
        if self.calculate_uncertainty == False:
            if self.evaluation == True:
                segmentation_regression_evaluation(self.path)

        return results,test_image_files, num_test_img

    def uncertainty_prediction(self, results,
                                checkpoint_filepath, 
                                network_input_size, 
                                Training_Input_shape, 
                                num_channels, 
                                test_image_files, 
                                num_test_img,
                                data_dimensions ,  
                                num_channels_label, 
                                Input_image_shape):    

        '''
        Start uncertainty prediction if selected for regression or semantic segmentation
        As suggested by Gal et. al.: https://arxiv.org/abs/1506.02142 
        And as implemented in: https://openreview.net/pdf?id=Sk_P2Q9sG
        Args:
            checkpoint_filepath: the directory where checkpoints are saved
            network_input_size: the dimensions of the input to the network
            Training_Input_shape: the shape of the images in the train dataset
            num_channels: number of channels (e.g.: 3 for RGB)
            test_image_files: list of filenames contained in the testset
            num_test_img: number of filenames in the testset
            data_dimensions: image dimensions
            num_channels_label: number of channels of the groundtruth (e.g.: 3 for RGB)
            Input_image_shape: The shape of the input images
        returns:
            Saves the results to the 'results' directory and the uncertainty estimations to the 'uncertainty' directory
        '''
        if data_dimensions == 3:
            logging.info("Using 3D UNet")
            if self.epochs > 0:
                uncertainty_weights = checkpoint_filepath
            else:
                uncertainty_weights = self.pretrained_weights
            model = UNetBuilder.unet3D(uncertainty_weights,
                                               network_input_size[-1],
                                               num_channels_label,
                                               loss_function = self.loss_function,
                                               num_classes = self.num_classes,
                                               Dropout_On=True)
        else:
            logging.info("Using 2D UNet")
            if self.epochs > 0:
                uncertainty_weights = checkpoint_filepath
            else:
                uncertainty_weights = self.pretrained_weights
            model = UNetBuilder.unet2D(uncertainty_weights,
                                               network_input_size[-1],
                                               num_channels_label,
                                               loss_function = self.loss_function,
                                               num_classes = self.num_classes,
                                               Dropout_On = True)
        resultsMCD = []
        for i in range(0, 20):
            testGene = testGenerator(Training_Input_shape, self.path, num_channels, test_image_files, self.use_algorithm)
            resultsMCD.append(model.predict_generator(testGene,
                                                              steps=num_test_img,
                                                              use_multiprocessing=False,
                                                              verbose=1))
        resultsMCD = np.array(resultsMCD)
        aleatoric_uncertainty = np.mean(resultsMCD * (1 - resultsMCD), axis=0)
        epistemic_uncertainty = np.mean(resultsMCD**2, axis = 0) - np.mean(resultsMCD, axis = 0)**2
        saveUncertainty(self.path + "/insights/", test_image_files, epistemic_uncertainty, aleatoric_uncertainty)
        uncertainty = epistemic_uncertainty + aleatoric_uncertainty
        saveResult(self.path + "/uncertainty/", test_image_files, uncertainty, Input_image_shape)
        if self.evaluation == True:
            segmentation_regression_evaluation(self.path)

    def run(self):    
        data_prepration_results = self.data_prepration()
        
        Training_Input_shape = data_prepration_results[0]
        num_channels = data_prepration_results[1]
        network_input_size = data_prepration_results[2]
        Input_image_shape = data_prepration_results[3]
        data_path = data_prepration_results[4]
        train_image_files = data_prepration_results[5]
        val_image_files = data_prepration_results[6] 
        steps_per_epoch = data_prepration_results[7]
        data_dimensions = data_prepration_results[8]
        val_image_files = data_prepration_results[9]

        TrainingDataGenerator, ValidationDataGenerator, num_channels_label = self.data_generator(   data_path, 
                                                                            Training_Input_shape, 
                                                                            num_channels, 
                                                                            train_image_files, 
                                                                            data_dimensions, 
                                                                            val_image_files)

        model = self.load_model( network_input_size,data_dimensions,num_channels_label)

        model, checkpoint_filepath = self.train_model(  model,
                                                        TrainingDataGenerator,
                                                        ValidationDataGenerator , 
                                                        steps_per_epoch, 
                                                        val_image_files  )

        results,test_image_files, num_test_img = self.test_set_evaluation( model, 
                                                                        Training_Input_shape, 
                                                                        num_channels,
                                                                        Input_image_shape)

        if self.calculate_uncertainty == True:
            self.uncertainty_prediction(    results,
                                            checkpoint_filepath,
                                            network_input_size,
                                            Training_Input_shape,
                                            num_channels,
                                            test_image_files,
                                            num_test_img,
                                            data_dimensions ,
                                            num_channels_label,
                                            Input_image_shape)
        model = None
    


'''
InstantDL
Data augmentations which can be executed on the fly
Written by Dominik Waibel
'''

import numpy as np
import skimage as sk
import random
from instantdl.data_generator.data import plot2images
from scipy.ndimage.interpolation import rotate
from scipy.ndimage.filters import gaussian_filter
import math
import os

def data_augentation(X, Y, data_gen_args, data_path_file_name):

    if "horizontal_flip" in data_gen_args:
        """ Flip image and groundtruth horizontally by a chance of 33% 
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            returns: two tensors, X, Y, which have the same dimensions as the input 
        """
        if data_gen_args["horizontal_flip"] == True & random.choice([True, False, False]) == True:
            X = np.flip(X, len(np.shape(X))-1)
            Y = np.flip(Y, len(np.shape(Y))-1)

    if "vertical_flip" in data_gen_args:
        """ Flip image and groundtruth vertically by a 33% chance
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            returns: two tensors, X, Y, which have the same dimensions as the input
        """
        if data_gen_args["vertical_flip"] == True & random.choice([True, False, False]) == True:
            X = np.flip(X, len(np.shape(X))-2)
            Y = np.flip(Y, len(np.shape(Y))-2)

    if "rotation_range" in data_gen_args and data_gen_args["rotation_range"] > 0:
        """ Rotate image and groundtruth by a random angle within the rotation range 
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            returns: two tensors, X, Y, which have the same dimensions as the input
        """
        angle =  np.random.choice(int(data_gen_args["rotation_range"]*100))/100
        if len(np.shape(X))-1 == 3:
            X = np.nan_to_num(rotate(X, angle, mode='nearest', axes=(2, 3), reshape=False))
            Y = np.nan_to_num(rotate(Y, angle, mode='nearest', axes=(2, 3), reshape=False))
        else:
            X = np.nan_to_num(rotate(X, angle, mode='nearest', axes=(1, 2), reshape=False))
            Y = np.nan_to_num(rotate(Y, angle, mode='nearest', axes=(1, 2), reshape=False))
        # Zoom so that there are no empty edges
        shape_X = np.shape(X)
        shape_Y = np.shape(Y)
        length_X = np.max(shape_X)
        zoom_length_X = length_X / (np.cos(math.radians(angle)) + np.sin(math.radians(angle)))
        zoom_length_Y = length_X / (np.cos(math.radians(angle)) + np.sin(math.radians(angle)))
        X = X[..., int((length_X - zoom_length_X) / 2) : int(length_X -  ((length_X - zoom_length_X) / 2)), int((length_X - zoom_length_X) / 2) : int(length_X - ((length_X - zoom_length_X) / 2)), :]
        X = np.nan_to_num(sk.transform.resize(X, shape_X))
        Y = Y[..., int((length_X - zoom_length_Y) / 2) : int(length_X -  ((length_X - zoom_length_Y) / 2)), int((length_X - zoom_length_Y) / 2) : int(length_X - ((length_X - zoom_length_Y) / 2)), :]
        Y = np.nan_to_num(sk.transform.resize(Y, shape_Y))


    if "width_shift_range" in data_gen_args and data_gen_args["width_shift_range"] > 0 & random.choice([True, False]) == True:
        """ Shift the image and groundtruth width by a number within the width shift range by a 50% chance
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            width shift range: float, amound of width shift
            returns: two tensors, X, Y, which have the same dimensions as the input
        """
        width_shift = np.random.choice(int(data_gen_args["width_shift_range"] * 100))
        shape_X = np.shape(X)
        shape_Y = np.shape(Y)
        if len(np.shape(X))-1 == 3:
            size = np.size(X[0, 0, :, 0, 0])
        else:
            size = np.size(X[0, 0, :, 0,])
        start_width_shift = np.random.choice(int(size - (size - width_shift)+1))

        X = X[:, :, start_width_shift: int(size - int(size * width_shift / 100)),...]
        X = np.nan_to_num(sk.transform.resize(X, shape_X))
        Y = Y[:, :, start_width_shift: int(size - int(size * width_shift / 100)),...]
        Y = np.nan_to_num(sk.transform.resize(Y, shape_Y))

    if "height_shift_range" in data_gen_args and data_gen_args["height_shift_range"] > 0 & random.choice([True, False]) == True:
        """ Shift the image and groundtruth height by a number within the width shift range by a 50% chance
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            height shift range: float, amound of height shift
            returns: two tensors, X, Y, which have the same dimensions as the input
        """
        height_shift = np.random.choice(int(data_gen_args["height_shift_range"] * 100))
        shape_X = np.shape(X)
        shape_Y = np.shape(Y)
        if len(np.shape(X))-1 == 3:
            size = np.size(X[0, 0, :, 0, 0])
        else:
            size = np.size(X[0, 0, :, 0,])
        start_heigth_shift = np.random.choice(int(size - (size - height_shift)+1))
        X = X[..., start_heigth_shift: int(size - int(size * height_shift / 100)), :]
        X = np.nan_to_num(sk.transform.resize(X, shape_X))
        Y = Y[..., start_heigth_shift: int(size - int(size * height_shift / 100)), :]
        Y = np.nan_to_num(sk.transform.resize(Y, shape_Y))

    if "zoom_range" in data_gen_args and data_gen_args["zoom_range"] > 0 & random.choice([True, False]) == True:
        """ Zooms the image and groundtruth to a random magnification in the zoom range and to a random position in the image by a 50% chance
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth
            zoom range: float, amound of zoom
            returns: two tensors, X, Y, which have the same dimensions as the input
        """
        zoom = np.random.choice(int(data_gen_args["zoom_range"] * 100)) + 1
        shape_X = np.shape(X)
        shape_Y = np.shape(Y)
        if len(np.shape(X))-1 == 3:
            size = np.size(X[0, 0, :, 0, 0])
        else:
            size = np.size(X[0, 0, :, 0,])
        x_position = np.random.choice(zoom)
        y_position = np.random.choice(zoom)
        X = X[..., x_position: int(x_position + size - zoom), y_position: int(y_position + size - zoom), :]
        X = np.nan_to_num(sk.transform.resize(X, shape_X))
        Y = Y[..., x_position:int(x_position + size - zoom), y_position: int(y_position + size - zoom), :]
        Y = np.nan_to_num(sk.transform.resize(Y, shape_Y))

    if "gaussian_noise" in data_gen_args and data_gen_args["gaussian_noise"] > 0 and random.choice([True, False, False]) == True:
        """ Adds gaussian noise to the image by a one-third chance
        # Arguments
            X: tensor, the input image
            gaussian_noise: float, amound of gaussian noise added
            returns: one tensors, X, which have the same dimensions as the input
        """
        value = data_gen_args["gaussian_noise"]
        X = X + np.random.normal(0, value)

    if "gaussian_blur_image" in data_gen_args and data_gen_args["gaussian_blur_image"] > 0 and random.choice([True, False, False]) == True:
        """ Blurs the input image in gaussian fashin by a 33% chance
        # Arguments
            X: tensor, the input image
            gaussian_blur: float, amound of gaussian blur added
            returns: one tensor, X, which have the same dimensions as the input
        """
        value = data_gen_args["gaussian_blur_image"]
        X = gaussian_filter(X, sigma=value)

    if "gaussian_blur_label" in data_gen_args and data_gen_args["gaussian_blur_label"] > 0 and random.choice([True, False, False]) == True:
        """ Blurs the groundtruth image in gaussian fashin by a 33% chance
        # Arguments
            Y: tensor, the groundtruth image
            gaussian_blur: float, amound of gaussian blur added
            returns: one tensor, Y, which have the same dimensions as the input
        """
        value = data_gen_args["gaussian_blur_label"]
        Y = gaussian_filter(Y, sigma=value)

    if "contrast_range" in data_gen_args and random.choice([True, False, False]) == True:
        """ Increases or decreases the contrast of the input in by the range given by a 33% chance
        # Arguments
            X: tensor, the groundtruth image
            contrast_range: float, amound of contrast added or removed
            returns: one tensor, X, which have the same dimensions as the input
        """
        range = np.random.uniform(-1,1) * data_gen_args["contrast_range"] + 1
        min_X = np.min(X)
        X = (X - min_X) * range + min_X

    if "brightness_range" in data_gen_args and random.choice([True, False]) == True:
        """ Increases or decreases the brightness of the input in by the range given by a 50% chance
        # Arguments
            X: tensor, the groundtruth image
            brightness_range: float, amound of gaussian noise added
            returns: one tensor, X, which have the same dimensions as the input
        """
        range = np.random.uniform(-1,1) * data_gen_args["brightness_range"]
        X = X + range * X

    if "threshold_background_image" in data_gen_args and random.choice([True, False, False]) == True:
        """ Thresholds the input image at the mean and sets every pixelvalue below the mean to zero by a 33% chance
        # Arguments
            X: tensor, the groundtruth image
            threshold_background_image: True or False
            returns: one tensor, X, which have the same dimensions as the input
        """
        mean_X = np.mean(X)
        X[X < mean_X] = 0

    if "threshold_background_groundtruth" in data_gen_args and random.choice([True, False, False]) == True :
        """ Thresholds the groundtruth image at the mean and sets every pixelvalue below the mean to zero by a 33% chance
        # Arguments
            Y: tensor, the groundtruth image
            threshold_background_groundtruth: True or False
            returns: one tensor, Y, which have the same dimensions as the input
        """
        mean_Y = np.mean(Y) * 0.8
        Y[Y < mean_Y] = 0

    if "binarize_mask" in data_gen_args and data_gen_args["binarize_mask"] == True:
        """ Binarize the groundtruth image at the mean and sets every pixelvalue below the mean to zero and every above to one
        # Arguments
            Y: tensor, the groundtruth image
            binarize_mask: True or False
            returns: one tensor, Y, which have the same dimensions as the input
        """
        mean_Y = np.mean(Y)
        Y[Y <= mean_Y] = 0
        Y[Y > mean_Y] = 1
        Y = np.nan_to_num(Y)

    if "save_augmented_images" in data_gen_args and data_gen_args["save_augmented_images"] == True:
        """ save all augmented images to a folder named "Augmentations" in the project folder to make augmentations trackable
        # Arguments
            X: tensor, the input image
            Y: tensor, the groundtruth image
            save_augmented_images: True or False
            returns: Saves the image pair as .png to the Augmentations folder
        """
        data_path, file_name = os.path.split(data_path_file_name)
        Aug_path = (data_path + '/Augmentations/' )
        os.makedirs(Aug_path, exist_ok=True)
        #TODO: Check if the folder and saving really works with the path
        title = file_name.split("'")[1]
        title = os.path.splitext(title)[0]
        #datacomb = np.concatenate((X, Y), axis=2).astype("uint8")
        #sk.io.imsave(Aug_path + title, datacomb[0,...])
        if len(np.shape(X)) == 5:
            plot2images(X[0, 10, ... ,0], Y[0, 10, :, :,0], Aug_path, title)
        elif len(np.shape(X)) == 4 and np.shape(X)[-1] == 3:
            plot2images(X[0, ...], Y[0,...], Aug_path, title)
        elif len(np.shape(X)) == 4 and np.shape(X)[-1] == 1:
            plot2images(X[0, ..., 0], Y[0, :, :, 0], Aug_path, title)
    #logging.info("Augmented Dimensions:", np.shape(X), np.shape(Y))
    return X, Y


'''
InstantDL
Written by Dominik Waibel and Ali Boushehri

In this file the functions are started to train and test the networks
'''

from instantdl.utils import *

class InstanceSegmentation(object):
    def __init__(   self,
                    use_algorithm,
                    path,
                    pretrained_weights = None,
                    batchsize = 2,
                    iterations_over_dataset = 100,
                    data_gen_args = None,
                    loss_function = "mse",
                    num_classes = 1,
                    image_size = None,
                    seeds=False,
                    calculate_uncertainty = False,
                    evaluation = True):

        self.use_algorithm = "InstanceSegmentation"
        self.path = path
        self.pretrained_weights = pretrained_weights
        self.batchsize = batchsize
        self.iterations_over_dataset = iterations_over_dataset
        self.loss_function = loss_function
        self.num_classes = num_classes
        self.image_size = image_size
        self.calculate_uncertainty = calculate_uncertainty
        self.evaluation = evaluation
        
        if data_gen_args is None:
            self.data_gen_args = dict()
        else:
            self.data_gen_args = data_gen_args
    


    def run(self):    
        '''
        Initialize a model for instance segmentation
        '''
        UseResnet = 50

        RESULTS_DIR = os.path.join(self.path, "results/")

        image_files = os.listdir(os.path.join(self.path + "/train"))

        '''
        Set values for config file
        '''
        RCNNConfig.NAME = "Image"
        RCNNConfig.classdefs = ("Image", self.num_classes, "Classes")
        RCNNConfig.BACKBONE = str("resnet" + str(UseResnet))
        print("RCNNConfig.BACKBONE", RCNNConfig.BACKBONE)
        RCNNConfig.STEPS_PER_EPOCH = int(len(image_files))
        RCNNConfig.VALIDATION_STEPS = 10
        RCNNConfig.NUM_CLASSES = 1 + self.num_classes # Background + Classes
        RCNNConfig.data_gen_args = self.data_gen_args

        lenval = len(image_files) * 0.2
        validation_spilt_id = np.array(list(range(0, len(image_files), int(len(image_files) / lenval))))
        print(validation_spilt_id)
        VAL_IMAGE_IDS = []
        for i in range(0, len(image_files)):
            if i in validation_spilt_id:
                VAL_IMAGE_IDS.append(image_files[i])

        trainsubset = "train"
        testsubset = "test"
        '''
        Initialize a model for instance segmentation with pretrained weights or imagenet weights
        '''
        if self.pretrained_weights == None:
            print("Weights == False")
            weights = "last"
        else:
            weights = self.pretrained_weights
        dataset = self.path
        logs = os.path.join(self.path, "logs")

        # Configurations
        config = RCNNConfig()
        config.display()

        # Create model
        model = RCNNmodel.MaskRCNN(mode="training", config=config, model_dir=logs)

        # Select weights file to load
        if os.path.isfile(str(self.pretrained_weights)):
            #if os.path.exists(self.pretrained_weights):
            print("Using pretrained weights from path")
            weights_path = weights
        elif self.pretrained_weights == "last":
            if os.path.exists(model.find_last()):
                print("Using last weights")
                weights_path = model.find_last()
        else:
            print("No weigths given: Using imagenet weights")
            weights_path = model.get_imagenet_weights()

        # Load weights
        print("Loading weights ", weights_path)
        if weights.lower() == "coco":
            # Exclude the last layers because they require a matching
            # number of classes
            model.load_weights(weights_path, by_name=True, exclude=[
                "mrcnn_class_logits", "mrcnn_bbox_fc",
                "mrcnn_bbox", "mrcnn_mask"])
        else:
            model.load_weights(weights_path, by_name=True)
        tensorboard = TensorBoard(log_dir="logs/" + self.path + "/" + format(time.time()))  # , update_freq='batch')
        custom_callbacks = [EarlyStopping(monitor='val_loss', min_delta=0, patience=5, verbose=0, mode='auto'), tensorboard]
        if self.iterations_over_dataset > 0:
            print("Start train")
            train(model, dataset, trainsubset, VAL_IMAGE_IDS, self.iterations_over_dataset, custom_callbacks)

        print("Iterations are set to:", self.iterations_over_dataset)
        print("Testing with weights:", weights_path)
        config = RCNNInferenceConfig()
        model = RCNNmodel.MaskRCNN(mode="inference", config=config, model_dir=logs)
        if self.iterations_over_dataset > 0:
            weights_path = model.find_last()
        model.load_weights(weights_path, by_name=True)
        test_Val_IMAGE_IDS = []
        detect(model, dataset, testsubset, RESULTS_DIR, test_Val_IMAGE_IDS)
        if self.evaluation == True:
            segmentation_regression_evaluation(self.path)

        model = None



import sys
import logging
import argparse
import os
import json
from instantdl.data_generator.data_generator import *
from instantdl.data_generator.auto_evaluation_classification import classification_evaluation
from instantdl.data_generator.auto_evaluation_segmentation_regression import segmentation_regression_evaluation
from instantdl.segmentation.UNet_models import UNetBuilder
from keras.callbacks import ModelCheckpoint, TensorBoard, EarlyStopping
import time
import tensorflow as tf
tf.get_logger().setLevel('WARNING')
from keras import backend as K
from keras.applications.resnet50 import ResNet50
from instantdl.classification.ResNet50 import ResNet50
from instantdl.segmentation.RCNNSettings import RCNNInferenceConfig, train, detect
import instantdl.segmentation.RCNNmodel as RCNNmodel
from instantdl.segmentation.RCNNSettings import RCNNConfig
tf.config.experimental.list_physical_devices('GPU')
from instantdl.classification.ResNet50 import ResNet50 #, get_imagenet_weights
import glob
from keras.optimizers import Adam, SGD
from instantdl.data_generator.data import write_logbook
logging.basicConfig(level=logging.INFO)
from instantdl.data_generator.metrics4losses import *

def load_json(file_path):
    with open(file_path, 'r') as stream:
        return json.load(stream)



from keras.models import *
from keras.layers import *
from keras.optimizers import *
import numpy as np
from keras.losses import *

def nn_base(input_tensor=None, trainable=False):


    input_shape = (None, None, 3)

    if input_tensor is None:
        img_input = Input(shape=input_shape)
    else:
        if not K.is_keras_tensor(input_tensor):
            img_input = Input(tensor=input_tensor, shape=input_shape)
        else:
            img_input = input_tensor

    bn_axis = 3

    # Block 1
    x = Conv2D(64, (3, 3), activation='relu', padding='same', name='block1_conv1')(img_input)
    x = Conv2D(64, (3, 3), activation='relu', padding='same', name='block1_conv2')(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='block1_pool')(x)

    # Block 2
    x = Conv2D(128, (3, 3), activation='relu', padding='same', name='block2_conv1')(x)
    x = Conv2D(128, (3, 3), activation='relu', padding='same', name='block2_conv2')(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='block2_pool')(x)

    # Block 3
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='block3_conv1')(x)
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='block3_conv2')(x)
    x = Conv2D(256, (3, 3), activation='relu', padding='same', name='block3_conv3')(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='block3_pool')(x)

    # Block 4
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block4_conv1')(x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block4_conv2')(x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block4_conv3')(x)
    x = MaxPooling2D((2, 2), strides=(2, 2), name='block4_pool')(x)

    # Block 5
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block5_conv1')(x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block5_conv2')(x)
    x = Conv2D(512, (3, 3), activation='relu', padding='same', name='block5_conv3')(x)
    # x = MaxPooling2D((2, 2), strides=(2, 2), name='block5_pool')(x)

    return x

def rpn_layer(base_layers, num_anchors):
    """Create a rpn layer
        Step1: Pass through the feature map from base layer to a 3x3 512 channels convolutional layer
                Keep the padding 'same' to preserve the feature map's size
        Step2: Pass the step1 to two (1,1) convolutional layer to replace the fully connected layer
                classification layer: num_anchors (9 in here) channels for 0, 1 sigmoid activation output
                regression layer: num_anchors*4 (36 in here) channels for computing the regression of bboxes with linear activation
    Args:
        base_layers: vgg in here
        num_anchors: 9 in here

    Returns:
        [x_class, x_regr, base_layers]
        x_class: classification for whether it's an object
        x_regr: bboxes regression
        base_layers: vgg in here
    """
    x = Conv2D(512, (3, 3), padding='same', activation='relu', kernel_initializer='normal', name='rpn_conv1')(base_layers)

    x_class = Conv2D(num_anchors, (1, 1), activation='sigmoid', kernel_initializer='uniform', name='rpn_out_class')(x)
    x_regr = Conv2D(num_anchors * 4, (1, 1), activation='linear', kernel_initializer='zero', name='rpn_out_regress')(x)

    return [x_class, x_regr, base_layers]

num_anchors = 9
shared_layers = nn_base(img_input)


# File for custom loss functions

from keras import backend as K
from keras.losses import binary_crossentropy
import tensorflow as tf

def dice_coef(y_true, y_pred, smooth=1):
    """
    Dice = (2*|X & Y|)/ (|X|+ |Y|)
         =  2*sum(|A*B|)/(sum(A^2)+sum(B^2))
    ref: https://arxiv.org/pdf/1606.04797v1.pdf
    """
    intersection = K.sum(K.abs(y_true * y_pred), axis=-1)
    return (2. * intersection + smooth) / (K.sum(K.square(y_true),-1) + K.sum(K.square(y_pred),-1) + smooth)

def dice_loss(y_true, y_pred):
    """ Dice loss function.
    Parameters
    ----------
    y_true : keras tensor
        tensor containing target mask.
    y_pred : keras tensor
        tensor containing predicted mask.
    Returns
    -------
    keras tensor
        tensor containing dice loss.
    """
    return 1-dice_coef(y_true, y_pred)# In this file loss function, which are not in keras are implemented

def dice_crossentropy_loss(y_true, y_pred):
    """ Dice loss + binary crossentropy function.
    Parameters
    ----------
    y_true : keras tensor
        tensor containing target mask.
    y_pred : keras tensor
        tensor containing predicted mask.
    Returns
    -------
    keras tensor
        tensor containing dice loss + binary crossentropy
    """
    if K.sum(y_true) != 0:
        vol_loss = K.abs(K.sum(y_true)-K.sum(y_pred))/(K.sum(y_true)*2)
    else:
        vol_loss = 0
    return dice_loss(y_true, y_pred) + binary_crossentropy(y_true, y_pred) + vol_loss

def tversky_loss(y_true, y_pred, alpha=0.3, beta=0.7, smooth=1e-10):
    """ Tversky loss function.
    Parameters
    ----------
    y_true : keras tensor
        tensor containing target mask.
    y_pred : keras tensor
        tensor containing predicted mask.
    alpha : float
        real value, weight of '0' class.
    beta : float
        real value, weight of '1' class.
    smooth : float
        small real value used for avoiding division by zero error.
    Returns
    -------
    keras tensor
        tensor containing tversky loss.
    """
    y_true = K.flatten(y_true)
    y_pred = K.flatten(y_pred)
    truepos = K.sum(y_true * y_pred)
    fp_and_fn = alpha * K.sum(y_pred * (1 - y_true)) + beta * K.sum((1 - y_pred) * y_true)
    answer = (truepos + smooth) / ((truepos + smooth) + fp_and_fn)
    return -answer

def binary_focal_loss_fixed(y_true, y_pred, gamma = 2., alpha = 0.25):
    """ Binary focal loss function.
    Parameters
    ----------
    y_true : keras tensor
        tensor containing target mask.
    y_pred : keras tensor
        tensor containing predicted mask.
    alpha : float
        real value, weight of '0' class.
    beta : float
        real value, weight of '1' class.
    smooth : float
        small real value used for avoiding division by zero error.
    Returns
    -------
    keras tensor
        tensor containing binary focal loss.
    """
    y_true = tf.cast(y_true, tf.float32)
    epsilon = K.epsilon()
    y_pred = K.clip(y_pred, epsilon, 1.0 - epsilon)
    p_t = tf.where(K.equal(y_true, 1), y_pred, 1 - y_pred)
    alpha_factor = K.ones_like(y_true) * alpha
    alpha_t = tf.where(K.equal(y_true, 1), alpha_factor, 1 - alpha_factor)
    cross_entropy = -K.log(p_t)
    weight = alpha_t * K.pow((1 - p_t), gamma)
    loss = weight * cross_entropy
    loss = K.mean(K.sum(loss, axis=1))
    return loss


'''
Author: Dominik Waibel

This file contains the functions with which the data is improted and exported before training and testing and after testing.
'''

from instantdl.data_generator.data import *
import numpy as np
from skimage.io import imread
from skimage.transform import resize
from instantdl.data_generator.data_augmentation import data_augentation
import os
import csv as csv
import sys
import copy
import glob
import logging
from keras.utils import to_categorical
from skimage.color import gray2rgb
import warnings

def get_min_max(data_path, folder_name, image_files):
    '''
    This function gets the minimum and maximum pixel values for the folder
    
    Args:
        data_path (str): path to project folder
        folder_name (str): one of the data folder names, e.g. image or groundtruth
        train_image_files (list): list of image files
    
    return: 
        min_value: the minimum pixel value of the dataset 
        max_value: the maximum pixel value of the dataset
    '''
    num_img = len(image_files)
    Xmin = np.empty(num_img)
    Xmax = np.empty(num_img)
    for i, img_file in enumerate(image_files):
        if img_file.endswith(".npy"):
            Xmin[i] = np.min(np.array(np.load(data_path + folder_name + img_file)))
            Xmax[i] = np.max(np.array(np.load(data_path + folder_name + img_file)))
        else:
            Xmin[i] = np.min(np.array(imread(data_path + folder_name + img_file)))
            Xmax[i] = np.max(np.array(imread(data_path + folder_name + img_file)))
    
    min_value = np.min(Xmin)
    max_value = np.max(Xmax)

    logging.info("min value of %s is %s and the max value is %s" % \
                                            (folder_name, min_value, max_value))
    return min_value, max_value

def import_image(path_name):
    '''
    This function loads the image from the specified path
    NOTE: The alpha channel is removed (if existing) for consistency

    Args:
        path_name (str): path to image file
    
    return: 
        image_data: numpy array containing the image data in at the given path. 
    '''
    if path_name.endswith('.npy'):
        image_data = np.array(np.load(path_name))
    else:
        image_data = imread(path_name)
        # If has an alpha channel, remove it for consistency
    if np.array(np.shape(image_data))[-1] == 4:
        image_data = image_data[:,:,0:3]
    return image_data


def image_generator(    Training_Input_shape, batchsize, num_channels, 
                        train_image_file, folder_name, data_path, 
                        X_min, X_max, use_algorithm):
    '''
    This function normalizes the imported images, resizes them and create batches

    Args:
        Training_Input_shape: The dimensions of one image used for training. Can be set in the config.json file
        batchsize: the batchsize used for training
        num_channels: the number of channels of one image. Typically 1 (grayscale) or 3 (rgb)
        train_image_file: the file name
        folder_name: the folder name of the file to be imported
        data_path: the project directory
        X_min: the minimum pixel value of this dataset
        X_max: the maximum pixel value of this dataset
    return: 
        X: a batch of image data with dimensions (batchsize, x-dim, y-dim, [z-dim], channels)
    '''
    X = []
    for i in range(0, batchsize):
        img_file = train_image_file[i]
        image_data = import_image(data_path + folder_name + img_file)
        if np.shape(image_data) != tuple(Training_Input_shape):
            #The Resizing fundtion changes the array values, therefore shift them back to the original range
            min_value = np.min(image_data)
            max_value = np.max(image_data)
            image_data = resize(image_data, Training_Input_shape)
            newmin = np.min(image_data)
            newmax = np.max(image_data)
            image_data = ((image_data - newmin) / (newmax - newmin)) * (max_value - min_value) + min_value
        image_data = (image_data - X_min) / (X_max - X_min)
        X.append(image_data)
    X = np.stack(X, axis = 0)
    if np.shape(X)[-1] != num_channels:
       X= X[..., np.newaxis]
    return X


def training_data_generator(Training_Input_shape, batchsize, num_channels, 
                            num_channels_label, train_image_files, 
                            data_gen_args, data_dimensions,data_path, use_algorithm):
    '''
    Generate the data for training and return images and groundtruth 
    for regression and segmentation

    Args
        Training_Input_shape: The dimensions of one image used for training. Can be set in the config.json file
        batchsize: the batchsize used for training
        num_channels: the number of channels of one image. Typically 1 (grayscale) or 3 (rgb)
        num_channels_label: the number of channels of one groundtruth image. Typically 1 (grayscale) or 3 (rgb)
        train_image_files: list of files in the training dataset
        data_gen_args: augmentation arguments
        data_dimensions: the dimensions of one image or the dimension to which the image should be resized
        data_path: path to the project directory
        use_algorithm: the selected network (UNet, ResNet50 or MRCNN)
    
    return: 
        X_train: batched training data
        Y: label or ground truth
    '''
    Folder_Names = ["/groundtruth/", "/image/", "/image1/", "/image2/", "/image3/", "/image4/", "/image5/", "/image6/", "/image7/"]
    X_min = np.zeros((len(Folder_Names)))
    X_max = np.zeros((len(Folder_Names)))
    for i, folder_name in enumerate(Folder_Names):
        if os.path.isdir(data_path + folder_name) == True:
            X_min[i], X_max[i] = get_min_max(data_path, folder_name, train_image_files)

    while True:
        def grouped(train_image_files, batchsize):
            return zip(*[iter(train_image_files)] * batchsize)
        for train_image_file in grouped(train_image_files, batchsize):
            for index, folder_name in enumerate(Folder_Names):
                if os.path.isdir(data_path + folder_name) == True:
                    if "groundtruth" in folder_name:
                        GT_Input_image_shape = np.array(copy.deepcopy(Training_Input_shape))
                        GT_Input_image_shape[-1] = num_channels_label
                        GT_Input_image_shape = tuple(GT_Input_image_shape)
                        if GT_Input_image_shape[-1] == 0:
                            GT_Input_image_shape = GT_Input_image_shape[0:-1]
                        Y = image_generator(GT_Input_image_shape, batchsize, num_channels_label, 
                                                train_image_file, folder_name, data_path, X_min[index], X_max[index], use_algorithm)
                    if "image" in folder_name and index > 1:
                        imp = image_generator(Training_Input_shape, batchsize, num_channels, train_image_file, 
                                                folder_name, data_path, X_min[index], X_max[index], use_algorithm)
                        X = np.concatenate([X, imp], axis = -1)
                    if "image" in folder_name and index == 1:
                        X = image_generator(Training_Input_shape, batchsize, num_channels, 
                        train_image_file, folder_name, data_path, X_min[index], X_max[index], use_algorithm)
            X_train, Y = data_augentation(X, Y, data_gen_args, data_path + str(train_image_file))
            X_train = np.nan_to_num(X_train)
            Y = np.nan_to_num(Y)
            yield (X_train, Y)


def training_data_generator_classification(Training_Input_shape, 
                                            batchsize, num_channels, 
                                            num_classes, train_image_files, 
                                            data_gen_args, data_path, use_algorithm):
    '''
    Generate the data for training and return images and groundtruth for classification

    Args
        Training_Input_shape: The dimensions of one image used for training. Can be set in the config.json file
        batchsize: the batchsize used for training
        num_channels: the number of channels of one image. Typically 1 (grayscale) or 3 (rgb)
        num_classes: the number of classes of the dataset, set in the config.json
        train_image_files: list of filenames in the training dataset
        data_gen_args: augmentation arguments
        data_path: path to the project directory
        use_algorithm: the selected network (UNet, ResNet50 or MRCNN)
    
    return: 
        X: batched training data
        Y: groundtruth labels
    '''
    csvfilepath = os.path.join(data_path + '/groundtruth/', 'groundtruth.csv')
    Folder_Names = ["/image/", "/image1/", "/image2/", "/image3/", "/image4/", "/image5/", "/image6/", "/image7/"]
    X_min = np.zeros((len(Folder_Names)))
    X_max = np.zeros((len(Folder_Names)))
    for i, folder_name in enumerate(Folder_Names):
        if os.path.isdir(data_path + folder_name) == True:
            X_min[i], X_max[i] = get_min_max(data_path, folder_name, train_image_files)
    logging.info("array of min values: %s" % X_min)
    logging.info("array of max values: %s" % X_max)
    while True:
        def grouped(train_image_files, batchsize):
            return zip(*[iter(train_image_files)] * batchsize)
        for train_image_file in grouped(train_image_files, batchsize):
            for index, folder_name in enumerate(Folder_Names):
                if os.path.isdir(data_path + folder_name) == True:
                    if "image" in folder_name and index > 0:
                        imp = image_generator(Training_Input_shape, batchsize, num_channels, 
                                                        train_image_file, folder_name, data_path, 
                                                        X_min[index], X_max[index], use_algorithm)
                                                        
                        X = np.concatenate([X, imp], axis = -1)
                    if "image" in folder_name and index == 0:
                        X = image_generator(Training_Input_shape, batchsize, num_channels, 
                        train_image_file, folder_name, data_path, X_min[index], X_max[index], use_algorithm)
            
            X, Trash = data_augentation(X, X, data_gen_args, data_path + str(train_image_file))
            label = np.zeros((batchsize))
            for j in range(0, batchsize):
                img_file = train_image_file[j]
                with open(csvfilepath) as csvfile:
                    reader = csv.DictReader(csvfile)
                    for row in reader:
                        if row['filename'] == img_file:
                            label[j] = row['groundtruth']
            label = to_categorical(label, num_classes)
            yield (X, label)

 
def testGenerator(Input_image_shape, path, num_channels, test_image_files, use_algorithm):
    '''
    Generate test images for segmentation, regression and classification

    Args
        Input_image_shape: The dimensions of one image used for training. Can be set in the config.json file
        path: path to the project directory
        num_channels: the number of channels of one image. Typically 1 (grayscale) or 3 (rgb)
        test_image_files: list of filenames in the test dataset
        use_algorithm: the selected network (UNet, ResNet50 or MRCNN)
    
    return: 
        X: one image on which a model prediction is executed
    '''
    test_path = path + "/test/"
    batchsize = 1
    Folder_Names = ["/image/", "/image1/", "/image2/", "/image3/", 
                        "/image4/", "/image5/", "/image6/", "/image7/"]
    X_min = np.zeros((len(Folder_Names)))
    X_max = np.zeros((len(Folder_Names)))
    for i, folder_name in enumerate(Folder_Names):
        if os.path.isdir(test_path + folder_name) == True:
            X_min[i], X_max[i] = get_min_max(test_path, folder_name, test_image_files)
    logging.info(test_image_files)
    logging.info("len test files %s" % len(test_image_files))
    while True:
        for test_file in test_image_files:
            test_file = [test_file]
            for index, folder_name in enumerate(Folder_Names):
                if os.path.isdir(test_path + folder_name) == True:
                    imp = image_generator(Input_image_shape, batchsize, num_channels, 
                                    test_file, folder_name, test_path, X_min[index], X_max[index], use_algorithm)
                    if index > 0 :
                        X = np.concatenate([X, imp], axis = -1)
                    else:
                        X = imp
            yield X


def saveUncertainty(path, test_image_files, epi_uncertainty, ali_uncertainty):
    '''
    saves the predicted uncertainty estimates to a .csv file
    Args:
        path: path to the project directory
        test_image_files: list of filenames in the test dataset
        results: the predicted segmentation or image
    return: None
    '''
    logging.info("Save result")
    os.makedirs(path, exist_ok=True)
    with open(path + 'uncertainty.csv', 'w') as writeFile:
        writer = csv.writer(writeFile)
        writer.writerow(['filename', 'summed uncertainty', 'mean epistemic uncertainty', 'median epistemic uncertainty', 'mean aleatoric uncertainty', 'median aleatoric uncertainty'])
        for i in range(0, len(epi_uncertainty)-1):
            writer.writerow([test_image_files[i], np.mean(epi_uncertainty[i,...])+np.mean(ali_uncertainty[i,...]), np.mean(epi_uncertainty[i,...]), np.median(epi_uncertainty[i,...]), np.mean(ali_uncertainty[i,...]), np.median(ali_uncertainty[i,...])])


def saveResult(path, test_image_files, results, Input_image_shape):
    '''
    saves the predicted segmentation or image to the Results 
                            folder in the project directory
    
    Args:
        path: path to the project directory
        test_image_files: list of filenames in the test dataset
        results: the predicted segmentation or image
        Input_image_shape: The dimensions of one image used for 
                            training. Can be set in the config.json file
    
    return: None
    '''
    results = results * 255.
    logging.info("Save result")
    os.makedirs(path, exist_ok=True)
    logging.info("shape npyfile %s" % (np.shape(results),))
    logging.info("test_image_files %s" % len(test_image_files))
    for i in range(len(test_image_files)):
        logging.info("filename %s" % test_image_files[i])
        titlenpy = (test_image_files[i] + "_predict")
        logging.info(i)
        if np.shape(results)[-1] == 0:
            results = results[:,:,:,0]
        minnpy = np.min(results[i, ...])
        maxnpy = np.max(results[i, ...])
        npy_resized = np.squeeze(resize(results[i, ...], Input_image_shape))
        newmin = np.min(npy_resized)
        newmax = np.max(npy_resized)
        npy_resized = ((npy_resized - newmin) / (newmax - newmin)) * (maxnpy - minnpy) + minnpy
        plottestimage_npy(npy_resized.astype("uint8"), path, titlenpy)

def saveResult_classification(path, test_image_files, results):
    '''
    saves a .csv file to the results folder in the project directory
            containing the predicted labels on the testset

        path: path to the project directory
        test_image_files: list of filenames in the test dataset
        results: list of predicted labels
    
    return: None
    '''
    logging.info("Save result")
    save_path = (path + '/results/')
    os.makedirs(path + "/results", exist_ok=True)
    with open(save_path + '/results.csv', 'w') as writeFile:
        writer = csv.writer(writeFile)
        writer.writerow(['filename', 'prediction', 'Probability for each possible outcome'])
        for i in range(0, len(results)-1):
            writer.writerow([test_image_files[i], np.argmax(results[i,...]), results[i,...]])

def saveResult_classification_uncertainty(path, test_image_files, results, 
                                                    MCprediction, combined_uncertainty):
    '''
    saves a .csv file to the results folder in the project directory
            containing the predicted labels on the testset
    
    Args:        
        path: path to the project directory
        test_image_files: list of filenames in the test dataset
        results: list of predicted labels
        MCprediction: list of predicted labels based on the likeliest prediction using MC Dropout
        combined_uncertainty: list of uncertaintys for each image
    
    return: None
    '''
    logging.info("Save result")
    save_path = (path + '/results/')
    os.makedirs(path + "/results", exist_ok=True)
    with open(save_path + 'results.csv', 'w') as writeFile:
        writer = csv.writer(writeFile)
        writer.writerow(['filename', 'prediction', 'Probability for each possible outcome',
                                    'MC Prediction', 'Certertainty: 0 is certain, high is uncertain'])
        for i in range(0, len(results)-1):
            writer.writerow([test_image_files[i], np.argmax(results[i,...]), \
                                results[i,...], MCprediction[i], np.abs(combined_uncertainty[i])])


def training_validation_data_split(data_path):
    '''
    Import filenames and split them into train and validation set according to the variable -validation_split = 20%
    Splits files in the train folder into a training and validation dataset and returns both lists containing the filenames
    
    Args
        
        data_path: path to the project directory
    
    return: 
        train_image_files: list of filenames of training data files
        val_image_files: list of filenames of validaton data files
    '''
    image_files = os.listdir(os.path.join(data_path + "/image"))
    lenval = int(len(image_files) * 0.2)
    validation_spilt_id = np.array(list(range(0, len(image_files), int(len(image_files) / lenval))))
    logging.info(validation_spilt_id)
    train_image_files = []
    val_image_files = []
    for i in range(0, len(image_files)):
        if i in validation_spilt_id:
            val_image_files.append(image_files[i])
        if i not in validation_spilt_id:
            train_image_files.append(image_files[i])
    train_image_files = np.random.permutation(train_image_files)
    logging.info("Found: %s images in training set" % len(train_image_files))
    logging.info("Found: %s images in validation set" % len(val_image_files))
    return train_image_files, val_image_files


def get_input_image_sizes(path, use_algorithm):
    '''
    Get the size of the input images and check dimensions

    Args:
        path: path to project directory
        use_algorithm: the selected network (UNet, ResNet50 or MRCNN)
    
    return: 
        Training_Input_shape: the shape of the training data
        num_channels: number of channels
        Input_image_shape: the shape of the input image
    '''
    data_path = path + '/train'
    img_file = os.listdir(data_path + "/image/")[0]
    Input_image_shape = np.array(np.shape(np.array(import_image(data_path + "/image/" + img_file))))
    logging.info("Input shape Input_image_shape %s" % Input_image_shape)
    if use_algorithm in ["Regression", "Segmentation"]:
        logging.info("Input_image_shape %s" % Input_image_shape)
        if int(Input_image_shape[0]) not in [int(16), int(32),int(64),int(128),int(256),int(512),int(1024),int(2048)]:
            if int(Input_image_shape[1]) not in [int(16), int(32),int(64),int(128),int(256),int(512),int(1024),int(2048)]:
                sys.exit("The Input data needs to be of pixel dimensions: 16, 32, 64, 128, 256, 512, 1024 or 2048 in each dimension")

    # If has an alpha channel, remove it for consistency
    Training_Input_shape = copy.deepcopy(Input_image_shape)
    if Training_Input_shape[-1] == 4:
        logging.info("Removing alpha channel")
        Training_Input_shape[-1] = 3

    if all([Input_image_shape[-1] != 1, Input_image_shape[-1] != 3]):
        logging.info("Adding an empty channel dimension to the image dimensions")
        Training_Input_shape = np.array(tuple(Input_image_shape) + (1,))

    if use_algorithm == "Classification":
        Training_Input_shape[-1] = 3

    num_channels = Training_Input_shape[-1]
    input_size = tuple(Training_Input_shape)
    logging.info("Input size is: %s" % (input_size,))
    return tuple(Training_Input_shape), num_channels, Input_image_shape



import numpy as np
import os
import sklearn
from sklearn.metrics import accuracy_score, confusion_matrix, roc_curve, roc_auc_score, auc
import numpy as np
from keras.utils import to_categorical
import pandas as pd
import matplotlib.pyplot as plt
import datetime
import csv
import logging

def get_auc(path, y_test, y_score, n_classes):
    '''
    calculates the area uncer curve and saves the AUC-curve to the insights folder
    :param path: path to directory
    :param y_test: groundtruth labels
    :param y_score: prediction labels
    :param n_classes: number of classes in classification task
    :return: Save figure to insights folder and return roc_auc values
    '''
    fpr = []
    tpr = []
    roc_auc = []
    for i in range(n_classes):
        fpr_out, tpr_out, _ = roc_curve(y_test[:, i].astype("float"), y_score[:, i].astype("float"))
        fpr.append(fpr_out)
        tpr.append(tpr_out)
        roc_auc.append(auc(fpr_out, tpr_out))
    plt.figure()
    for i in range(n_classes):
        plt.plot(fpr[i], tpr[i], label= "Class:" + str(i) +' ROC curve (area = %0.2f)' % roc_auc[i])
        plt.plot([0, 1], [0, 1], 'k--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic example')
    plt.legend(loc='bottom right')
    plt.savefig(path + "/insights/AUC.png")
    return roc_auc

def get_confusion_matrix(path, Groundtruth, Results):
    '''
    :param path: path to project directory
    :param Groundtruth: groundtruth class labels
    :param Results: predicted class labels
    :return: safe confusion matrix to insight folder
    '''
    confusion_matr = confusion_matrix(Groundtruth, Results)
    normalized_confustion_matrix = confusion_matr.astype('float') / (confusion_matr.sum(axis=1)[:, np.newaxis]+10e-10)
#    logging.info("The confusion matrix is: \n", confusion_matrix)
    cmap = plt.cm.Blues
    title = 'Confusion matrix'
    fig, ax = plt.subplots()
    im = plt.imshow(normalized_confustion_matrix * 100, interpolation='nearest', cmap='binary')
    cbar = plt.colorbar(im)
    cbar.ax.tick_params(labelsize=15)
    plt.xticks(np.arange(confusion_matr.shape[0]), fontsize=15, rotation=90)
    plt.yticks(np.arange(confusion_matr.shape[1]), fontsize=15)
    plt.title(title, fontsize=15)
    plt.ylabel('True Class', fontsize=15)
    plt.xlabel('Predicted Class', fontsize=15)
    plt.tight_layout()
    plt.savefig(path + "/insights/Confusion_Matrix.png") #TODO Plot cuts legend when saving

def load_data(path):
    '''
    imports the predictions and groundtruth and creates a table called Predictions_and_Results_Table in the insights folder
    containing the predictions and results for each filename
    :param path: path to project directory
    :return: the imported groundtruth, results and sigmoid outputs as lists
    '''
    Groundtruth_in = pd.read_csv(path + "/test/groundtruth/groundtruth.csv")
    Results_in = pd.read_csv(path + "/results/results.csv")
    Results = []
    Groundtruth = []
    Sigmoid_output = []
    Sigmoid_max_output = []
    os.makedirs(path + '/insights/', exist_ok = True)
    with open(path + '/insights/Predictions_and_Results_Table.csv', 'w') as writeFile:
        writer = csv.writer(writeFile)
        writer.writerow(['filename', 'prediction', 'groundtruth'])
        for i in range(len(Results_in)):
            loc = Groundtruth_in.loc[Groundtruth_in['filename'] == Results_in["filename"][i]]
            if str(loc['filename'].values)[2:-2] == Results_in["filename"][i]:
                Results.append(Results_in["prediction"][i])
                Groundtruth.append((loc["groundtruth"]).values.astype(int))
                Sigmoid_output_i = Results_in["Probability for each possible outcome"][i]
                Sigmoid_max_output.append(max(Sigmoid_output_i))
                Sigmoid_output_i = np.array((Sigmoid_output_i)[1:-1].split())
                Sigmoid_output.append(Sigmoid_output_i)
                writer.writerow([Results_in["filename"][i], Results_in["prediction"][i], (loc["groundtruth"].values)])
    Groundtruth = np.array(Groundtruth)
    Results = np.array(Results)
    Sigmoid_output = np.array(Sigmoid_output)
    return Groundtruth, Results, Sigmoid_output

def classification_evaluation(path):
    '''
    opens the results and groundtruth and caluclates the accuracy and area under curve

    :param path: path to the project directory
    '''
    Groundtruth, Results, Sigmoid_output = load_data(path)
    accuracy = accuracy_score(Groundtruth, Results)
    Sigmoid_output = np.squeeze(Sigmoid_output)
    logging.info("The accuracy score is: %f" % accuracy)
    n_classes = len(np.unique(Groundtruth))
    GT = to_categorical(Groundtruth)
    os.makedirs((path + "/insights/"), exist_ok=True)
    roc_auc = get_auc(path, GT, Sigmoid_output, int(n_classes))
    weigted_roc_auc = 0
    for index, i in enumerate(np.unique(Groundtruth)):
        logging.info("Groundtruth.count(i) %d %d" % (i, np.count_nonzero(Groundtruth == i)))
        weigted_roc_auc += np.count_nonzero(Groundtruth == i)*roc_auc[index]
    weigted_roc_auc /= len(Groundtruth)
    logging.info("The weigted roc auc is: %f" % weigted_roc_auc)
    get_confusion_matrix(path, Groundtruth, Results)
    f = open(path + '/insights/Evaluation_results.txt', 'a')
    f.write('\n')
    f.write('\n' + "Evaluated at: " + str(datetime.datetime.now())[:16])
    f.write('\n' + "Accuracy : " + str(accuracy))
    for i in range(len(roc_auc)):
        f.write('\n' + "Area under curve class " + str(i) + ": " + str(roc_auc[i]))
    f.close()


"""
InstantDL
Utils for data evaluation
Written by Dominik Waibel
"""

import numpy as np
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Computer Modern Roman"
from skimage import io
from skimage.io import imread
from skimage.color import gray2rgb, rgb2gray
from skimage.transform import resize
import os
import logging

def normalize(data):
	'''
	:param data: data to be normalized0
	:return: normalize data between 0 and 1
	'''
	mindata = np.min(data)
	maxdata = np.max(data)
	data = (data - mindata) / (maxdata - mindata)
	data = np.nan_to_num(data)
	return data

def import_images(import_dir, files, new_file_ending):
	'''
	:param import_dir: directory from where the images are imported
	:param files: list of filenames
	:param new_file_ending: file ending to be attached to the imagenames
	:return: returns a stack containing the image data and their names in the correct order
	'''
	data = []
	names = []
	for file in files:
		if new_file_ending is not None:
			file = (file + new_file_ending)
		if file.endswith(".npy"):
			imp = np.array(np.load(os.path.join(import_dir, file)))
		else:
			imp = np.array(imread(os.path.join(import_dir, file)))
		if np.shape(imp)[-1] == 1:
			imp = imp[...,-1]
		if np.shape(imp)[-1] is not 3:
			imp = gray2rgb(imp)
		data.append(imp)
		names.append(file)
	data = np.array(data)
	logging.info("Checking dimensions:")
	if np.shape(data)[-1] == 1:
		data = data[...,0]
	logging.info("Shape of data imported is:")
	logging.info(np.shape(data))
	return data, names

def calcerrormap(prediction, groundtruth):
	'''
	:param prediction: stack of images in the prediction dataset
	:param groundtruth: stack of images in the groundtruth dataset
	:return: stack with the absolute and relative differene between the prediction and groundtruth
	'''
	logging.info(np.shape(groundtruth))
	logging.info(np.shape(prediction))
	groundtruth = np.asarray(groundtruth, dtype=float)
	prediction = np.asarray(prediction, dtype=float)
	groundtruth_norm = (groundtruth - np.mean(groundtruth))/(np.std(groundtruth))
	prediction_norm = (prediction - np.mean(prediction))/(np.std(prediction))
	groundtruth_fs = (groundtruth - np.min(groundtruth))/(np.max(groundtruth)- np.min(groundtruth))
	prediction_fs = (prediction - np.min(prediction))/(np.max(prediction)- np.min(prediction))
	abs_errormap_norm = ((groundtruth_fs - prediction_fs))
	rel_errormap_norm = np.abs(np.divide(abs_errormap_norm, groundtruth_norm, out=np.zeros_like(abs_errormap_norm), where=groundtruth_norm!=0))
	rel_error_norm = np.mean(np.concatenate(np.concatenate((rel_errormap_norm))))
	logging.info("The relative Error over the normalized dataset is:",rel_error_norm, " best ist close to zero")
	return abs_errormap_norm, rel_errormap_norm

def prepare_data_for_evaluation(root_dir, max_images):
	'''
	:param root_dir: path to directory
	:param max_images: the maximum number of images to to be evaluated
	:return: executes the quantitative and visual asssesment of the model predition
			saves quantitative results to the insights folder and visual results to the evaluation folder
	'''
	test_dir = root_dir + "/test/"
	results_dir = root_dir + "/results/"
	report_dir = root_dir + "/evaluation/"
	insights_dir = root_dir + "/insights/"
	if os.path.isdir(test_dir + "/image/") == True:
		RCNN = False
	else:
		RCNN = True
	#Set results dir only for RCNN
	results_dir = root_dir + "/results/"
	groundtruth_exists = False
	if os.path.exists(root_dir + "/test/groundtruth/") and os.path.isdir(root_dir + "/test/groundtruth/"):
		groundtruth_exists = True
	if RCNN == True:
		logging.info("Resizing MaskRCNN images to 256, 256 dimensins to make them stackable")
		image_fnames = os.listdir(test_dir)
		image_fnames = image_fnames[0:max_images]
		image = []
		groundtruth = []
		predictions = []
		logging.info("importing", len(image_fnames), "files")
		for name in image_fnames:
			image_folder = (test_dir+name+"/image/"+name+".png")
			imagein = (np.array(rgb2gray(io.imread(image_folder))))
			image.append(resize(imagein, (256,256)))
			groundtruths_imp = []
			for file in os.listdir(test_dir + name + "/mask/"):
				groundtruths_in = np.array(io.imread(test_dir + name + "/mask/" + file))
				groundtruths_in = resize(rgb2gray(groundtruths_in), (256, 256))
				groundtruths_imp.append(groundtruths_in)
			groundtruth.append(np.sum(groundtruths_imp, axis=0))
			prediction_folder = results_dir + name + ".npy"
			prediction = np.load(prediction_folder)
			prediction = np.sum(np.where(prediction == 1, 1, 0), axis = -1)
			prediction[prediction > 1] = 1
			predictions.append(resize(prediction, (256,256)))
		logging.info("pred %s" % np.shape(predictions))
		logging.info("gt %s" % np.shape(groundtruth))
		predictions = np.array(predictions)
		groundtruth = np.array(groundtruth)
		abs_errormap_norm, rel_errormap_norm = calcerrormap(predictions, groundtruth)

	else:
		image_files = os.listdir(test_dir + "/image/")
		if max_images is not None:
			image_files = image_files[0: max_images]
		image, image_fnames = import_images(test_dir + "/image/", image_files, None)
		predictions, _ = import_images(results_dir, image_files, "_predict.tif")
		if os.path.isdir(test_dir + "/groundtruth/"):
			groundtruth, _ = import_images(test_dir + "/groundtruth/", image_files, None)
		image, image_fnames = import_images(test_dir + "/image/", image_files, None)
		if os.path.isdir(test_dir + "image2"):
			image2, _ = import_images(test_dir + "image2", image_files, None)
			image1, _ = import_images(test_dir + "image1", image_files, None)
		elif os.path.isdir(test_dir + "image1"):
			image1, _ = import_images(test_dir + "image1", image_files, None)
	if os.path.isdir(root_dir + "uncertainty"): 
		uncertainty, _ = import_images(root_dir + "uncertainty", image_files, "_predict.tif")
	if os.path.isdir(test_dir + "/groundtruth/"):
		abs_errormap_norm, rel_errormap_norm = calcerrormap(predictions, groundtruth)

	os.makedirs(root_dir + "/insights/", exist_ok=True)
	np.save(root_dir + "/insights/" + "image", image)
	np.save(root_dir + "/insights/" + "prediction", predictions)
	if os.path.isdir(test_dir + "/groundtruth/") and len(groundtruth) > 1:
		np.save(root_dir + "/insights/" + "groundtruth", groundtruth)
		np.save(root_dir + "/insights/" + "abs_errormap", abs_errormap_norm)
		np.save(root_dir + "/insights/" + "rel_errormap", rel_errormap_norm)
	np.save(root_dir + "/insights/" + "image_names", image_fnames)
	if os.path.isdir(root_dir + "uncertainty"):
		np.save(root_dir + "/insights/" + "uncertainty", uncertainty/255.)
	if os.path.isdir(test_dir + "image1"):
		logging.info("Two")
		np.save(root_dir + "/insights/" + "image1", image1)
	if os.path.isdir(test_dir + "image2"):
		logging.info("Three")
		np.save(root_dir + "/insights/" + "image1", image1)
		np.save(root_dir + "/insights/" + "image2", image2)



'''
Import the dependencies
'''
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Computer Modern Roman"
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Computer Modern Roman"
import os
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams["font.family"] = "Computer Modern Roman"
import copy
from matplotlib.colorbar import Colorbar
from skimage.color import rgb2gray, gray2rgb
import logging
from sklearn.metrics import jaccard_score
from sklearn.metrics import roc_curve, auc
from instantdl.evaluation.Utils_data_evaluation import prepare_data_for_evaluation
from skimage import filters
import datetime

def threshold(img):
    img[img < np.mean(img)] = 0
    return img

def binarize(data):
    threshold = filters.threshold_otsu(data)
    data[data <= threshold] = 0
    data[data > threshold] = 1
    return data

def normalize(data):
	'''
	:param data: data to be normalized0
	:return: normalize data between 0 and 1
	'''
	mindata = np.min(data)
	maxdata = np.max(data)
	data = (data - mindata) / (maxdata - mindata)
	data = np.nan_to_num(data)
	return data

def AUC(pred, gt):
    '''
    :param data: groundtruth and prediction
    :return: Area under curve
    '''
    ground_truth_labels = gt.flatten() # we want to make them into vectors
    score_value = pred.flatten() # we want to make them into vectors
    fpr, tpr, _ = roc_curve(ground_truth_labels,score_value)
    roc_auc = auc(fpr,tpr)
    return roc_auc

def getPearson(gt, pred):
    '''
    :param data: groundtruth and prediction
    :return: pearson correlation
    '''
    pearson_all = [np.zeros(np.shape(gt)[0])]
    for index in range(len(pearson_all)):
        gtp = np.array(gt[index].flatten())
        predp = np.array(pred[index].flatten())
        pearson_all[index] = np.mean(np.corrcoef(gtp, predp))
        print(pearson_all)
    pearson = np.mean(pearson_all)
    return pearson, pearson_all

def save_2Dimages(data, names, z_index, path_name, image_name):
    logging.info("plot len data")
    logging.info(len(data))
    plt.figure(figsize = (len(data)*8,10))
    grid = plt.GridSpec(4, len(data), height_ratios=[0.01,0.3,0.02,0.01], width_ratios=len(data) * [1])
    logging.info("shape data save")
    logging.info(np.shape(data))
    for img, image in enumerate(data):
        shape = np.shape(image)
        #logging.info("shape plot image", np.shape(image))
        plts = plt.subplot(grid[0:2, img])
        if names[img] == "abs_errormap":
            max_val = np.max(np.abs(image))
            pltcberror = plt.title("errormap", size = 50)
            pltcberror = plt.imshow(rgb2gray(image), cmap='seismic', vmin = -max_val, vmax = max_val)
        elif names[img] == "uncertainty":
            plts = plt.title(names[img] + str(np.mean(image)), size = 50)
            max_val = np.max(np.abs(image))
            pltcbunc = plt.title("uncertainty", size = 50)
            pltcbunc = plt.imshow(np.max(image)-rgb2gray(image), cmap='Greys', vmin = -max_val, vmax = max_val)
        else:
            plts = plt.imshow(rgb2gray(image), cmap='Greys')
            plts = plt.title(names[img], size = 50)
            pltz = plt.xlabel("x (Pixel)", size=50)
        if img == 0:
            pltz = plt.ylabel("y (Pixel)", size=50)

        plts = plt.tick_params(labelsize=50)
        if img > 0:
            plts = plt.tick_params(
            axis='y',
            labelleft = False)
            plts = plt.tick_params(labelsize=50)
        if names[img] == "abs_errormap":
            logging.info("colorbar")
            if "uncertainty" in names:
                cbax = plt.subplot(grid[2:3,-2])
            else:
                cbax = plt.subplot(grid[2:3,-1])
            cbax = Colorbar(ax = cbax, mappable = pltcberror,orientation="horizontal")
            cbax.set_ticks([])
            cbax.set_label('Over-     Under- \n prediction', size = 50)
        if names[img] == "uncertainty":
            logging.info("colorbar uncertanty")
            cbay = plt.subplot(grid[2:3,-1])
            cbay = Colorbar(ax = cbay,  mappable = pltcbunc, orientation="horizontal")
            cbay.set_ticks([])
            cbay.set_label('Low         High', size = 50)
        plt.savefig(os.path.join(path_name, z_index + image_name + ".png"), dpi=50,bbox_inches='tight')
    plt.show()
    plt.close()


def save_3Dimages(data, names, z_index, path_name, image_name):
    plt.figure(figsize=(len(data) * 8, 12))
    logging.info("plot len data", len(data))
    grid = plt.GridSpec(5, len(data), height_ratios=[0.01, 0.3, 0.1, 0.02, 0.01], width_ratios=len(data) * [1])

    logging.info("shape data save", np.shape(data))
    for img, image in enumerate(data):
        shape = np.shape(image)
        logging.info("shape image", np.shape(image))
        x_image = image[int(shape[0] / 2), ...]
        z_image = image[:, int(shape[1] / 2), :]
        logging.info("shape plot image", np.shape(x_image))
        plts = plt.subplot(grid[0:2, img])
        plts = plt.plot(np.arange(np.shape(x_image)[1]), [np.shape(x_image)[1] / 2] * np.shape(x_image)[1], color='0.5')
        plts = plt.title(names[img], size=50)
        plts = plt.tick_params(labelsize=50)
        plts = plt.xticks([])
        if img > 0:
            plts = plt.yticks([])
        if img == 0:
            plts = plt.ylabel("y (Pixel)", size=50)
            plts = plt.yticks(size=40)
        if names[img] == "abs_errormap":
            max_val = np.max(np.abs(x_image))
            pltcberror = plt.imshow(rgb2gray(x_image), cmap='seismic', vmin=-max_val, vmax=max_val)
            pltcberror = plt.subplot(grid[2, img])
            pltcberror = plt.plot(np.arange(np.shape(z_image)[1]), [np.shape(z_image)[0] / 2] * np.shape(z_image)[1],
                                  color='0.5')
        elif names[img] == "uncertainty":
            max_val = np.max(np.abs(x_image))
            pltcbunc = plt.title("uncertainty", size=50)
            pltcunc = plt.imshow(rgb2gray(x_image), cmap='Greys', vmin=-max_val, vmax=max_val)
            pltcunc = plt.subplot(grid[2, img])
            pltcunc = plt.plot(np.arange(np.shape(z_image)[1]), [np.shape(z_image)[0] / 2] * np.shape(z_image)[1],
                               color='0.5')
        else:
            pltxz = plt.imshow(x_image, cmap='Greys')
            pltxz = plt.subplot(grid[2, img])
            pltxz = plt.plot(np.arange(np.shape(z_image)[1]), [np.shape(z_image)[0] / 2] * np.shape(z_image)[1],
                             color='0.5')
        if img > 0:
            pltxy = plt.tick_params(axis='y', labelleft=False)
            pltxy = plt.tick_params(labelsize=40)
            pltxz = plt.yticks([])
        if img == 0:
            pltxz = plt.ylabel("y (Pixel)", size=50)
        if names[img] == "abs_errormap":
            max_val = np.max(np.abs(x_image))
            pltcerror = plt.imshow(rgb2gray(z_image), cmap='seismic', vmin=-max_val, vmax=max_val)
            pltcberror = plt.yticks([])
        elif names[img] == "uncertainty":
            max_val = np.max(np.abs(x_image))
            pltcunc = plt.imshow(rgb2gray(np.max(z_image) - z_image), cmap='Greys', vmin=-max_val, vmax=max_val)
        else:
            pltz = plt.imshow(z_image, cmap='Greys')
            pltz = plt.tick_params(labelsize=40)
            pltz = plt.xlabel("x (Pixel)", size=50)
        if img == 0:
            pltz = plt.ylabel("z (Pixel)", size=50)
        if img > 0:
            pltz = plt.tick_params(axis='y', labelleft=False, labelsize=40)
        # pltz = plt.tick_params(labelsize=40)

    if "abs_errormap" in names:
        if "uncertainty" in names:
            cbax = plt.subplot(grid[3:4, -2])
        else:
            cbax = plt.subplot(grid[3:4, -1])
        cbax = Colorbar(ax=cbax, mappable=pltcerror, orientation="horizontal")
        cbax.set_ticks([])
        cbax.set_label('Over-  Under- \n prediction', size=50)
        cbax.ax.tick_params(labelsize=50)
    if "uncertainty" in names:
        cbax = plt.subplot(grid[3:4, -1])
        cbax = Colorbar(ax=cbax, mappable=pltcunc, orientation="horizontal")
        cbax.set_ticks([])
        cbax.set_label('Low      High', size=50)
        cbax.ax.tick_params(labelsize=50)
    logging.info(path_name + z_index  + image_name + ".png")
    plt.savefig(path_name + z_index  + image_name + ".png", dpi=50, bbox_inches='tight')
    plt.show()
    plt.close()


def quantitative_evaluation(path, data, names):
    groundtruth_norm = (data[names.index("groundtruth")] - np.min(data[names.index("groundtruth")])) / (np.max(data[names.index("groundtruth")]) - np.min(data[names.index("groundtruth")]))
    prediction_norm = (data[names.index("prediction")] - np.min(data[names.index("prediction")])) / (np.max(data[names.index("prediction")]) - np.min(data[names.index("prediction")]))
    rel_errormap_norm = np.abs(np.divide(data[names.index("abs_errormap")], groundtruth_norm, out=np.zeros_like(data[names.index("abs_errormap")]),
                                         where=groundtruth_norm != 0))

    groundtruth_binary = binarize(data[names.index("groundtruth")])
    prediction_binary = binarize(data[names.index("prediction")])
    groundtruth_norm = normalize(data[names.index("groundtruth")])
    prediction_norm = normalize(data[names.index("prediction")])
    Pearson, Pearson_all = getPearson(data[names.index("prediction")], data[names.index("groundtruth")])

    logging.info("Metriccs:")
    logging.info("The mean absolute error on the normalized dataset is: " + str(np.abs(np.mean(data[names.index("abs_errormap")]))))
    logging.info("The Pearson coefficient is: " + str(np.abs(Pearson)))
    logging.info("The Jaccard index is: " + str(jaccard_score(prediction_binary.flatten(), groundtruth_binary.flatten())))
    #logging.info("The AUC is:", AUC(prediction_binary.flatten(), groundtruth_binary.flatten()))
    # logging.info("The Information score is: ", mutual_info_score(np.concatenate(np.concatenate(prediction_norm)), np.concatenate(np.concatenate(groundtruth_norm))))
    # logging.info("The rand score is:" , adjusted_rand_score(np.concatenate(np.concatenate(groundtruth_norm)), np.concatenate(np.concatenate(prediction_norm))))
    f = open(path + '/Evaluation_results.txt', 'a')
    f.write('\n' + "Evaluated at: " + str(datetime.datetime.now())[:16])
    f.write('\n' + "The mean absolute error on the normalized dataset is: " + str(np.abs(np.mean(data[names.index("abs_errormap")]))))
    f.write('\n' + "The Pearson coefficient is: " + str(np.abs(Pearson)))
    f.write('\n' + "The Jaccard index is: " + str(jaccard_score(prediction_binary.flatten(), groundtruth_binary.flatten())))
    #f.write('\n' + "The AUC is:" + str(AUC(prediction_binary.flatten(), groundtruth_binary.flatten())))
    f.close()

def visual_assesment(path, data, names):
    image_names = np.load(path + "/insights/image_names.npy")

    '''Function to execute image generation and saving'''
    os.makedirs(path + "/evaluation/", exist_ok=True)

    # Threshold the uncertainty map to obtain more meaningful images
    #if "uncertainty" in names:
    #    data[names.index("uncertainty")] = threshold(data[names.index("uncertainty")])

    for index, image in enumerate(data):
        if names[index] not in ["abs_errormap"]:
            if names[index] not in ["uncertainty"]:
                logging.info(index)
                logging.info(np.mean(np.mean(image)))
                logging.info(np.shape(image))
                data[index, ...] = normalize(data[index, ...])

    names_out = copy.deepcopy(names)
    report_dir = path + "/evaluation/"
    logging.info("Length of data")
    logging.info(len(data))
    if len(np.shape(data[0])) == 3 or len(np.shape(data[0])) == 4 and np.shape(data[0])[-1] == 3:
        for index in range(len(data[0])):
            logging.info("index")
            logging.info(index)
            logging.info("saving 2D image")
            '''Taking the slices in z-dimension one after another'''
            save_2Dimages(np.array(data)[:, index, ...], names, str(index), report_dir, image_names[index])

    if len(np.shape(data[0])) == 5 or len(np.shape(data[0])) == 5 and np.shape(data[0])[-1] == 4:
        for index in range(len(data[1])):
            logging.info(index)
            logging.info("saving 3D image")
            '''Taking the slices in z-dimension one after another'''
            logging.info("data shape", np.shape(np.array(data)))
            logging.info("save 3d", report_dir)
            save_3Dimages(np.array(data)[:, index, ...], names, str(index), report_dir, image_names[index])


def segmentation_regression_evaluation(path):
    savepath = path + "/insights/"
    max_images = 200000
    prepare_data_for_evaluation(path, max_images)
    # Load data from numpy stacks
    data_names = ["image", "image1", "image2", "image3", "prediction", "groundtruth", "abs_errormap", "uncertainty"]
    datain = []
    names = []
    for name in data_names:
        if os.path.isfile(path + "/insights/" + name + ".npy"):
            import_data = np.array(np.load(path + "/insights/" + name + ".npy").astype('float'))
            logging.info(name)
            logging.info(np.shape(import_data))
            datain.append(import_data)
            names.append(str(name))
    data = np.array(datain)
    image_names = np.load(path + "/insights/image_names.npy")
    logging.info(names)
    logging.info(np.shape(data))
    if "groundtruth" in names:
        quantitative_evaluation(savepath, data, names)
    else:
        logging.info("No Groundtruth given, therefore the quantitative performance can not be evaluated")
    visual_assesment(path, data, names)



'''
InstantDL
2D and 3D UNet models
Written by Dominik Waibel
The original publication is: arXiv:1505.04597
'''

from keras.models import *
from keras.layers import Input, Conv2D, Conv3D, LeakyReLU, BatchNormalization, Dropout, MaxPooling2D, concatenate, MaxPooling3D, UpSampling2D, UpSampling3D
from keras.optimizers import Adam, SGD, RMSprop
import numpy as np
from keras.losses import *
import logging

class UNetBuilder(object):
    '''
    Builds a 2D or 3D U-Net for image segmentation following the publication from Ronneberger et al.: https://arxiv.org/abs/1505.04597
    The layers consist of a convolution followed by a LeackyReLu activation and a Batch Normalization
    Dropout is integrated with a dropout rate of 20% during training to increase generalizability
    Padding has bee set to "same"
    '''

    def unet2D(pretrained_weights, num_channels_input, num_channels_label, num_classes, loss_function, Dropout_On, base_n_filters = 32):
        logging.info("started UNet")
        inputs = Input(shape = (None, None, num_channels_input))
        conv1 = Conv2D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(inputs)
        conv1 = BatchNormalization()(conv1)
        conv1 = LeakyReLU(alpha=0.2)(conv1)
        logging.info("finished 1. convolution")
        conv1 = Conv2D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(conv1)
        conv1 = BatchNormalization()(conv1)
        conv1 = LeakyReLU(alpha=0.2)(conv1)
        drop1 = Dropout(0.2)(conv1, training = Dropout_On)
        pool1 = MaxPooling2D(pool_size=(2, 2))(drop1)

        conv2 = Conv2D(base_n_filters*2, 3, padding='same', kernel_initializer='he_normal')(pool1)
        conv2 = BatchNormalization()(conv2)
        conv2 = LeakyReLU(alpha=0.2)(conv2)
        conv2 = Conv2D(base_n_filters*2, 3, padding='same', kernel_initializer='he_normal')(conv2)
        conv2 = BatchNormalization()(conv2)
        conv2 = LeakyReLU(alpha=0.2)(conv2)
        drop2 = Dropout(0.2)(conv2, training = Dropout_On)
        pool2 = MaxPooling2D(pool_size=(2, 2))(drop2)

        conv3 = Conv2D(base_n_filters*4, 3, padding='same', kernel_initializer='he_normal')(pool2)
        conv3 = BatchNormalization()(conv3)
        conv3 = LeakyReLU(alpha=0.2)(conv3)
        conv3 = Conv2D(base_n_filters*4, 3,  padding='same', kernel_initializer='he_normal')(conv3)
        conv3 = BatchNormalization()(conv3)
        conv3 = BatchNormalization()(conv3)
        conv3 = LeakyReLU(alpha=0.2)(conv3)
        drop3 = Dropout(0.2)(conv3, training = Dropout_On)
        pool3 = MaxPooling2D(pool_size=(2, 2))(drop3)

        conv4 = Conv2D(base_n_filters*8, 3, padding='same', kernel_initializer='he_normal')(pool3)
        conv4 = BatchNormalization()(conv4)
        conv4 = LeakyReLU(alpha=0.2)(conv4)
        conv4 = Conv2D(base_n_filters*8, 3, padding='same', kernel_initializer='he_normal')(conv4)
        conv4 = BatchNormalization()(conv4)
        conv4 = LeakyReLU(alpha=0.2)(conv4)
        drop4 = Dropout(0.2)(conv4, training = Dropout_On)
        pool4 = MaxPooling2D(pool_size=(2, 2))(drop4)

        conv5 = Conv2D(base_n_filters*16, 3, padding='same', kernel_initializer='he_normal')(pool4)
        conv5 = BatchNormalization()(conv5)
        conv5 = LeakyReLU(alpha=0.2)(conv5)
        conv5 = Conv2D(base_n_filters*16, 3, padding='same', kernel_initializer='he_normal')(conv5)
        conv5 = BatchNormalization()(conv5)
        conv5 = LeakyReLU(alpha=0.2)(conv5)
        drop5 = Dropout(0.2)(conv5, training = Dropout_On)

        up6 = Conv2D(base_n_filters*8, 2, padding='same', kernel_initializer='he_normal')(UpSampling2D(size=(2, 2))(drop5))
        merge6 = concatenate([drop4, up6], axis=3)
        conv6 = Conv2D(base_n_filters*8, 3,activation = 'relu',  padding='same', kernel_initializer='he_normal')(merge6)
        conv6 = BatchNormalization()(conv6)
        conv6 = LeakyReLU(alpha=0.1)(conv6)
        conv6 = Conv2D(base_n_filters*8, 3, activation = 'relu',  padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = BatchNormalization()(conv6)
        conv6 = LeakyReLU(alpha=0.1)(conv6)

        up7 = Conv2D(base_n_filters*4, 2, padding='same', kernel_initializer='he_normal')(UpSampling2D(size=(2, 2))(conv6))
        merge7 = concatenate([drop3, up7], axis=3)
        conv7 = Conv2D(base_n_filters*4, 3, padding='same', kernel_initializer='he_normal')(merge7)
        conv7 = BatchNormalization()(conv7)
        conv7 = LeakyReLU(alpha=0.1)(conv7)
        conv7 = Conv2D(base_n_filters*4, 3, padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = BatchNormalization()(conv7)
        conv7 = LeakyReLU(alpha=0.1)(conv7)

        up8 = Conv2D(base_n_filters*2, 2, padding='same', kernel_initializer='he_normal')(UpSampling2D(size=(2, 2))(conv7))
        merge8 = concatenate([drop2, up8], axis=3)
        conv8 = Conv2D(base_n_filters*2, 3, padding='same', kernel_initializer='he_normal')(merge8)
        conv8 = BatchNormalization()(conv8)
        conv8 = LeakyReLU(alpha=0.1)(conv8)
        conv8 = Conv2D(base_n_filters*2, 3, padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = BatchNormalization()(conv8)
        conv8 = LeakyReLU(alpha=0.1)(conv8)

        up9 = Conv2D(base_n_filters, 2, padding='same', kernel_initializer='he_normal')(UpSampling2D(size=(2, 2))(conv8))
        merge9 = concatenate([drop1, up9], axis=3)

        conv9 = Conv2D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(merge9)
        conv9 = BatchNormalization()(conv9)
        conv9 = LeakyReLU(alpha=0.1)(conv9)
        conv9 = Conv2D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = BatchNormalization()(conv9)
        conv9 = Conv2D(base_n_filters, 3,padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = BatchNormalization()(conv9)
        conv9 = LeakyReLU(alpha=0.1)(conv9)

        if num_classes > 1:
            conv10 = Conv2D(num_classes, (1), activation='softmax')(conv9)  # MultiClass segmentation with one-hot encoded image
        else:
            conv10 = Conv2D(num_channels_label, 1, activation='sigmoid')(conv9)  # Simple segmentation with only one label

        model2D = Model(inputs=inputs, outputs=conv10)
        logging.info("shape input UNet %s" % np.shape(inputs))
        logging.info("shape output UNet %s" % np.shape(conv10))
        model2D.compile(optimizer="Adam", loss = loss_function, metrics=['mse'])

        if (pretrained_weights):
            model2D.load_weights(pretrained_weights, by_name=True, skip_mismatch=True)
            for layer, pre in zip(model2D.layers, pretrained_weights):
                weights = layer.get_weights()
                if weights:
                    if np.array_equal(weights[0], pre[0]):
                        logging.info('not loaded %s' % layer.name)
                    else:
                        logging.info('loaded %s' % layer.name)

        return model2D

    def unet3D(pretrained_weights, num_channels_input, num_channels_label, num_classes, loss_function, Dropout_On, base_n_filters=32):
        logging.info("started UNet")
        inputs = Input(shape=(None, None, None, num_channels_input))
        conv1 = Conv3D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(inputs)
        conv1 = BatchNormalization()(conv1)
        conv1 = LeakyReLU(alpha=0.2)(conv1)
        logging.info("finished 1. convolution")
        conv1 = Conv3D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(conv1)
        conv1 = BatchNormalization()(conv1)
        conv1 = LeakyReLU(alpha=0.2)(conv1)
        drop1 = Dropout(rate=0.3)(conv1, training = Dropout_On)
        pool1 = MaxPooling3D(pool_size=(2, 2, 2))(drop1)

        conv2 = Conv3D(base_n_filters * 2, 3, padding='same', kernel_initializer='he_normal')(pool1)
        conv2 = BatchNormalization()(conv2)
        conv2 = LeakyReLU(alpha=0.2)(conv2)
        conv2 = Conv3D(base_n_filters * 2, 3, padding='same', kernel_initializer='he_normal')(conv2)
        conv2 = BatchNormalization()(conv2)
        conv2 = LeakyReLU(alpha=0.2)(conv2)
        drop2 = Dropout(rate=0.3)(conv2, training = Dropout_On)
        pool2 = MaxPooling3D(pool_size=(2, 2, 2))(drop2)

        conv3 = Conv3D(base_n_filters * 4, 3, padding='same', kernel_initializer='he_normal')(pool2)
        conv3 = BatchNormalization()(conv3)
        conv3 = LeakyReLU(alpha=0.2)(conv3)
        conv3 = Conv3D(base_n_filters * 4, 3, padding='same', kernel_initializer='he_normal')(conv3)
        conv3 = BatchNormalization()(conv3)
        conv3 = LeakyReLU(alpha=0.2)(conv3)
        drop3 = Dropout(rate=0.3)(conv3, training = Dropout_On)
        pool3 = MaxPooling3D(pool_size=(2, 2, 2))(drop3)

        conv4 = Conv3D(base_n_filters * 8, 3, padding='same', kernel_initializer='he_normal')(pool3)
        conv4 = BatchNormalization()(conv4)
        conv4 = LeakyReLU(alpha=0.2)(conv4)
        conv4 = Conv3D(base_n_filters * 8, 3, padding='same', kernel_initializer='he_normal')(conv4)
        conv4 = BatchNormalization()(conv4)
        conv4 = LeakyReLU(alpha=0.2)(conv4)
        drop4 = Dropout(rate=0.3)(conv4, training = Dropout_On)
        pool4 = MaxPooling3D(pool_size=(2, 2, 2))(drop4)

        conv5 = Conv3D(base_n_filters * 16, 3, padding='same', kernel_initializer='he_normal')(pool4)
        conv5 = BatchNormalization()(conv5)
        conv5 = LeakyReLU(alpha=0.2)(conv5)
        conv5 = Conv3D(base_n_filters * 16, 3, padding='same', kernel_initializer='he_normal')(conv5)
        conv5 = BatchNormalization()(conv5)
        conv5 = LeakyReLU(alpha=0.2)(conv5)
        drop5 = Dropout(rate=0.3)(conv5, training = Dropout_On)

        up6 = Conv3D(base_n_filters * 8, 2, padding='same', kernel_initializer='he_normal')(
            UpSampling3D(size=(2, 2, 2))(drop5))
        merge6 = concatenate([drop4, up6], axis=4)
        conv6 = Conv3D(base_n_filters * 8, 3, activation='relu', padding='same', kernel_initializer='he_normal')(merge6)
        conv6 = BatchNormalization()(conv6)
        conv6 = LeakyReLU(alpha=0.1)(conv6)
        conv6 = Conv3D(base_n_filters * 8, 3, activation='relu', padding='same', kernel_initializer='he_normal')(conv6)
        conv6 = BatchNormalization()(conv6)
        conv6 = LeakyReLU(alpha=0.2)(conv6)

        up7 = Conv3D(base_n_filters * 4, 2, padding='same', kernel_initializer='he_normal')(
            UpSampling3D(size=(2, 2, 2))(conv6))
        merge7 = concatenate([drop3, up7], axis=4)
        conv7 = Conv3D(base_n_filters * 4, 3, padding='same', kernel_initializer='he_normal')(merge7)
        conv7 = BatchNormalization()(conv7)
        conv7 = LeakyReLU(alpha=0.1)(conv7)
        conv7 = Conv3D(base_n_filters * 4, 3, padding='same', kernel_initializer='he_normal')(conv7)
        conv7 = BatchNormalization()(conv7)
        conv7 = LeakyReLU(alpha=0.1)(conv7)

        up8 = Conv3D(base_n_filters * 2, 2, padding='same', kernel_initializer='he_normal')(
            UpSampling3D(size=(2, 2, 2))(conv7))
        merge8 = concatenate([drop2, up8], axis=4)
        conv8 = Conv3D(base_n_filters * 2, 3, padding='same', kernel_initializer='he_normal')(merge8)
        conv8 = BatchNormalization()(conv8)
        conv8 = LeakyReLU(alpha=0.1)(conv8)
        conv8 = Conv3D(base_n_filters * 2, 3, padding='same', kernel_initializer='he_normal')(conv8)
        conv8 = BatchNormalization()(conv8)
        conv8 = LeakyReLU(alpha=0.1)(conv8)

        up9 = Conv3D(base_n_filters, 2, padding='same', kernel_initializer='he_normal')(
            UpSampling3D(size=(2, 2, 2))(conv8))
        merge9 = concatenate([drop1, up9], axis=4)
        conv9 = Conv3D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(merge9)
        conv9 = BatchNormalization()(conv9)
        conv9 = LeakyReLU(alpha=0.1)(conv9)
        conv9 = Conv3D(base_n_filters, 3, padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = BatchNormalization()(conv9)
        conv9 = LeakyReLU(alpha=0.1)(conv9)

        conv9 = Conv3D(3, 3, padding='same', kernel_initializer='he_normal')(conv9)
        conv9 = BatchNormalization()(conv9)
        conv9 = LeakyReLU(alpha=0.1)(conv9)

        if num_classes > 1:
            conv10 = Conv3D(num_classes, (1), activation='softmax')(conv9)  # Changed activiation from Relu to linear
        else:
            conv10 = Conv3D(num_channels_label, 1, activation='sigmoid')(conv9)  # Changed activiation from Relu to linear

        model3D = Model(inputs=inputs, outputs=conv10)
        logging.info("shape input UNet %s" % np.shape(inputs))
        logging.info("shape output UNet %s" % np.shape(conv10))
        model3D.compile(optimizer="Adam", loss=loss_function, metrics=['mse'])

        if (pretrained_weights):
            model3D.load_weights(pretrained_weights, by_name=True, skip_mismatch=True)
            for layer, pre in zip(model3D.layers, pretrained_weights):
                weights = layer.get_weights()
                if weights:
                    if np.array_equal(weights[0], pre[0]):
                        logging.info('not loaded %s' % layer.name)
                    else:
                        logging.info('loaded %s' % layer.name)
        return model3D


from instantdl.utils import *
from instantdl.classification.classification import Classification
from instantdl.segmentation.Regression import Regression
from instantdl.segmentation.InstanceSegmentation import InstanceSegmentation
from instantdl.segmentation.SemanticSegmentation import SemanticSegmentation

def GetPipeLine(configs):
    if "seeds" in configs:
        if configs["seeds"] == True:
            import numpy as np
            np.random.seed(123)
            import random as python_random
            python_random.seed(123)
            import tensorflow as tf
            tf.random.set_random_seed(123)
            sess = tf.Session(graph=tf.get_default_graph())
            K.set_session(sess)
    if configs["use_algorithm"] == "Classification":
        pipeline = Classification(**configs)
        return pipeline
    elif configs["use_algorithm"] == "Regression":
        pipeline = Regression(**configs)
        return pipeline
    elif configs["use_algorithm"] == "SemanticSegmentation" :
        pipeline = SemanticSegmentation(**configs)
        return pipeline
    elif configs["use_algorithm"] == "InstanceSegmentation":
        pipeline = InstanceSegmentation(**configs)
        return pipeline
    else: 
        raise KeyError("The use_algorithm should be set correctly")


import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='instantdl',
    version='1.0.5',
    description='An easy and convenient Deep Learning pipeline for image segmentation and classification',
    author='Dominik Waibel, Ali Boushehri',
    author_email='dominik.waibel@helmholtz-muenchen.de, ali.boushehri@roche.com',
    license='MIT',
    keywords='Computational Biology Deep Learning',
    url='https://github.com/marrlab/InstantDL',
    packages=find_packages(exclude=['doc*', 'test*']),
    install_requires=[  'keras>=2.2.4',
                        'tensorboard>=1.13.0,<=1.15.2'],
    long_description=read('README.md'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Image Recognition',
        'License :: OSI Approved :: MIT License',
    ],
)



'''
InstantDL
Written by Dominik Waibel and Ali Boushehri

In this file the functions are started to train and test the networks
'''

import os
import argparse
from instantdl.utils import load_json
from instantdl.data_generator.metrics4losses import *
from instantdl import GetPipeLine
import logging
from keras import backend as K

def start_learning(configs):
    logging.info("Start learning")
    logging.info(configs["use_algorithm"])

    pipeline = GetPipeLine(configs)

    pipeline.run()
    K.clear_session()


if __name__ == "__main__":

    parser = argparse.ArgumentParser( \
                            description='Starting the deep learning code')
    parser.add_argument('-c',\
                        '--config', \
                        default="config.json", \
                        help='config json file address', \
                        type=str)

    args = vars(parser.parse_args())

    configs = load_json(args['config'])

    for k in configs:
        logging.info("%s : %s \n" % (k,configs[k]))

   
    '''
    Sanity checks in order to ensure all settings in config
    have been set so the programm is able to run
    '''
    assert configs["use_algorithm"] in ['SemanticSegmentation',
                                        'Regression',
                                        'InstanceSegmentation',
                                        'Classification']

    if "batchsize" in configs:
        if not isinstance(configs["batchsize"], int):
            logging.warning("Batchsize has not been set. Setting batchsize = 1")
            batchsize = 2
    else:
        logging.warning("Batchsize has not been set. Setting batchsize = 1")
        configs["batchsize"] = 2

    if "iterations_over_dataset" in configs:
        if not isinstance(configs["iterations_over_dataset"], int):
            logging.warning("Epochs has not been set. Setting epochs = 500 and using early stopping")
            iterations_over_dataset = 100

    if "pretrained_weights" in configs:
        if not isinstance(configs["pretrained_weights"], str):
            if not os.path.isfile((configs["pretrained_weights"])):
                configs["pretrained_weights"] = None
    else:
        configs["pretrained_weights"] = None

    if configs["loss_function"] in ["dice_loss", "dice loss", "dice"]:
        configs["loss_function"] = dice_loss

    if configs["loss_function"] in ["tversky_loss", "tversky loss", "tversky"] :
        configs["loss_function"] = tversky_loss

    if configs["loss_function"] in ["focal_loss", "focal loss", "focal"] :
        configs["loss_function"] = binary_focal_loss_fixed

    if configs["loss_function"] == "binary crossentropy dice loss":
        configs["loss_function"] = dice_crossentropy_loss

    start_learning(configs)


from __future__ import print_function
import matplotlib.pylab as plt
import numpy as np
import os
from skimage.io import imsave
import datetime
import logging

def plottestimage_npy(image, path, title):
    '''
    saves the networks prediction to the results folder

    Args
        image: the networks prediction
        path: results path to where the image is saved
        title: image name
    
    return: None
    '''
    logging.info("shape of image %s" % (np.shape(image),))
    imsave(path + title + ".tif", image)

def plot2images(image, mask, path, title):
    '''
    saves a pair of images (image and mask) to directory

    Args   
        image: image data
        mask: mask data
        path: path to where the image pair is saved
        title: image name
    
    return: None
    '''
    plt.figure()
    plt.subplot(1, 2, 1)
    plt.imshow(image)
    plt.title("Image")
    plt.subplot(1, 2, 2)
    if len(np.shape(mask)) > 2:
        mask = mask[:,:,0]
    plt.imshow(mask)
    plt.title("Groundtruth")
    plt.savefig(os.path.join(path, str(title) + ".png"), dpi=50)
    plt.tight_layout()
    plt.close()


def write_logbook(path, epochs, loss, data_gen_args):
    f = open(path + '/Logbook.txt', 'a')
    f.write('\n' + "Run started at: " + str(datetime.datetime.now())[:16])
    f.write('\n' + "With lossfunction: " + str(loss) + " for : " + str(epochs) + " epochs")
    f.write('\n' + "The augmentations are: " + str(data_gen_args))
    f.write('\n')
    f.close()
